%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%     SYNTH: a transformation system for functional logic programs        %%
%%              based on needed narrowing                                  %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- use_module(library(system)).
:- use_module(library(lists)).


:- dynamic data_dec/1.
:- dynamic last_program/1,last_rule/1,rule/3,elim/2,elim_dt/3,unf/3,fold/3.
:- dynamic last_prg_vble/1,selected_program/1,selected_rule/1,conditions/0.
:- dynamic fold/0,elim_rule/0,deep/1,auto/1.
:- dynamic dt/3, strategy/1.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%                             PARSER                              %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

parser("",_) :- !.
parser(W,N) :- 
	retractall(parsWh(_)),
	assertz(parsWh(0)),
	sym_tab(ST),
	parser(W,N,ST).

parser("",_,ST) :- !, retractall(sym_tab(_)),assertz(sym_tab(ST)).

parser(W,N,ST) :-
	p_Rule(W,N,ST,ST2,S,M), 
	elim_space(S,W2),
	( [10|S2]=W2,
	  ! , elim_space(S2,W3),
	  parser(W3,M,ST2)
	; [] =S2
	).	

p_Rule("",_,_,_,_,_) :- !.
p_Rule(W,N,ST,ST_,S_,N):- % data
	append("data",S,W),
	elim_space(S,W2),
	p_TypeDecl(W2,ST,ST_,S_,DD),
	retract(data_dec(D)),
	append(D,[10|DD],D2),
	assert(data_dec(D2)).
p_Rule(W,N,ST,ST,S_,M):-
	p_Equat(W,ST,S_,Lhs,Rhs),
	M is N+1,
	assertz(rule(N,Lhs,Rhs)),
 	untranslate_rule(Lhs,Rhs),nl.

p_Equat(W,ST,S_,Lhs,Rhs) :-
	p_FunLHS(W,ST,S1,Lhs),
	elim_space(S1,W1),
	p_Rhs(W1,ST,S_,Rhs).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%                          Left Hand Side                         %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


p_FunLHS(W,ST,S_,apply(f(Name),Args)) :-	% "FunctionName {Pattern}"
	p_FunName(W,S,Name),
	elim_space(S,W2),
	p_ListPatt(W2,[],ST,_,S_,Args). 

p_FunLHS(W,ST,S_, apply(f(InfixOpId),[Exp1,Exp2])) :-	% "Pattern InfixOpId Pattern"
	p_Patt(W,ST,_,S,Exp1),
	elim_space(S,W2),
	p_InfixOpId(W2,S2,InfixOpId),
	elim_space(S2,W3),
	p_Patt(W3,ST,_,S_,Exp2).	

p_FunName([40|S],S_,Op) :-	% "'(' InfixOpId ')'"	40 = (
	elim_space(S,W),
	p_InfixOpId(W,S2,Op),
	elim_space(S2,W2),
	W2=[41|S_].		% 41 = )

p_FunName(W,S_,Id) :-		% "FunctionId"
	p_FunId(W,S_,Id).

p_ListPatt(W,AccPatt,ST,ST_,S_,LPatt) :-	% "...{Pattern}"
	p_Patt(W,ST,ST2,S2,Patt),	% Legge il Pattern e lo accumula 
	! , append(AccPatt,[Patt],AccPatt2),
	elim_space(S2,W2),
	p_ListPatt(W2,AccPatt2,ST2,ST_,S_,LPatt).

p_ListPatt(W,LPatt,ST,ST,W,LPatt).	% Finiti i Pattern: p_Patt fallito

p_Patt([40|S],ST,ST,S_,apply(c("()"),[])) :-	% "'(' ')'", con spazi in mezzo
	elim_space(S,W2) ,
	W2=[41|S_] .

p_Patt([40|S],ST,ST_,S_,Arg) :-			% "'(' DataConstrId ..."
	elim_space(S,W),
	p_DataConstrId(W,S2,Id),
	verify_DCID(Id,ST,ST2),
	elim_space(S2,W2),
	Transl_DataConstr= apply(c(Id),[]),
	p_DTerm_Tuple_List(W2,Transl_DataConstr,ST2,ST_,S_,Arg).

p_Patt([40|S],ST,ST_,S_,Arg) :-			% "'(' Pattern ..." 
	elim_space(S,W),
	p_Patt(W,ST,ST2,S2,El),			% e Pattern non e' DataConstrId
	elim_space(S2,W2) ,
	p_Tuple_List(W2,El,ST2,ST_,S_,Arg).

p_Patt([91|S],ST,ST,S_, apply(c("nil"),[])) :-	% "'['  ']'"
	elim_space(S,W) ,
	W=[93|S_].

p_Patt([91|S],ST,ST_,S_, List):- 		% "'[' Pattern {',' Pattern}']'"
	elim_space(S,W) ,
	p_Patt(W,ST,ST2,S2,Head),
	elim_space(S2,W2),
	p_SeqPatt(W2,Head,[93],c("cons"),ST2,ST_,S_,List).

p_Patt(W,ST,ST,S_,v(Id)) :-			% "VariableId"
	p_VarId(W,S_,Id).

p_Patt(W,ST,ST,S_,apply(c("true"),[])) :-	% "true"
	append("true",S_,W).
p_Patt(W,ST,ST,S_,apply(c("false"),[])):-	% "false"
	append("false",S_,W).

p_Patt(W,ST,ST,S_,Num) :-			% "Int"
	p_Int(W,S_,Num).

p_Patt(W,ST,ST_,S_,Arg) :-			% "DataConstrId"
	p_DataConstrId(W,S_,Id),
	Arg= apply(c(Id),[]),
	verify_DCID(Id,ST,ST_).

p_DTerm_Tuple_List(W, apply(c(DataConstrId),_),ST,ST_,S_, apply(c(DataConstrId),LArg)) :-
	W=[A|_],
	A \== 44 , A \== 58 , ! ,	% 44 = ,  58 = :
	p_Patt(W,ST,ST2,S,Patt),
	elim_space(S,W2),
	p_ListPatt(W2,[Patt],ST2,ST_,S2,LArg),
	elim_space(S2,W3),
	W3=[41|S_].			% 41 = )

p_DTerm_Tuple_List(W,Transl_DataConstrId,ST,ST_,S_,Arg) :-
	p_Tuple_List(W,Transl_DataConstrId,ST,ST_,S_,Arg).

p_Tuple_List([58|S],Head,ST,ST_,S_, apply(c("cons"), [Head,Tail])) :- 
					% "... ':' Pattern ')'"
	elim_space(S,W),		% 58 = :
	p_Patt(W,ST,ST_,S2,Tail),
	elim_space(S2,W2),
	W2=[41|S_].			% 41 = )

p_Tuple_List([44|S],Patt1,ST,ST_,S_, apply(c("()"),[Patt1,Rest]) ) :- 
					% "... ',' Pattern..." per le tuple
	elim_space(S,W),		% 44 = ,
	p_Patt(W,ST,ST2,S2,Patt2),
	elim_space(S2,W2),
	p_SeqPatt(W2,Patt2, [41],c("()"),ST2,ST_,S_,Rest).

p_SeqPatt([44|S],Patt1,End,F,ST,ST_,S_, apply(F, [Patt1,Tail])):-	
					% "... { ',' Pattern  ..." di una lista
	elim_space(S,W),		% 44 = ,
	p_Patt(W,ST,ST2,S2,Patt2),
	elim_space(S2,W2),
	p_SeqPatt(W2,Patt2,End,F,ST2,ST_,S_,Tail).

p_SeqPatt([93|S],Head,[93],_,ST,ST,S, 
		apply(c("cons"),[Head,apply(c("nil"),[])])).
					% " ... ']'' di una lista 93 = ]

p_SeqPatt([41|S],Head,[41],_,ST,ST,S, Head).	% " ... ')'' di una lista 93 = )

p_FunId([A|S],S_,Id) :- 
	p_Lowercase(A),
	p_TailId(S,[A],S_,Id).

p_VarId([A|S],S_,Id) :- 		
	(p_Uppercase(A) ; A==95),	% prima lettera maiuscola o underscore
	p_TailId(S,[A],S_,Id).
	
p_DataConstrId([A|S],S_,Id) :-
	p_Lowercase(A),			% prima lettera minuscola
	p_TailId(S,[A],S_,Id).

p_TailId([A|S],AccId,S_,Id) :-		
	p_correct(A), ! ,
	append(AccId,[A],AccId2), 
	p_TailId(S,AccId2,S_,Id).

p_TailId(S,Id,S,Id).

p_InfixOpId([A|S],S_,[39|Op2]) :-
	A==39 , 			% 39 = '
	! ,elim_space(S,W),
	p_FunId(W,S2,Op),
	append(Op,[39],Op2),
	elim_space(S2,W2),
	W2=[39|S_].

p_InfixOpId([A|S],S_,Op) :-
	(A==33;(A>=35,A=<38);A==42;A==43;(A>=45,A=<47);(A>=60,A=<64);
	A==94;A==124;A==126),
	! , p_TailInOpId(S,[A],S_,Op).

p_TailInOpId([A|S],AccOp,S_,Op) :-
	(A==33;(A>=35,A=<38);A==42;A==43;(A>=45,A=<47);(A>=60,A=<64);
	A==94;A==124;A==126), ! ,
	append(AccOp,[A],AccOp2),
	p_TailInOpId(S,AccOp2,S_,Op).

p_TailInOpId(S_,Op,S_,Op).

p_Int([A|S],S_,Num):-
	( A==45,
	  ! , [B|S2]=S, p_number(B), AccNum=[A,B]
	; p_number(A), AccNum = [A], S=S2
	),
	p_TailInt(S2,AccNum,S_,Num).

p_TailInt([N|S],AccI,S_,Num) :-
	p_number(N), ! ,
	append(AccI,[N],AccI2),
	p_TailInt(S,AccI2,S_,Num).

p_TailInt(S_,Num,S_,apply(c(Num),[])).


p_correct(A) :- p_letter(A) ; p_number(A) ; A == 95 .	% 95 = _

p_letter(A) :- p_Lowercase(A) ; p_Uppercase(A).

p_Lowercase(A) :- 97 =< A, A =< 122. % 97 = 'a' .. 122 = 'z'

p_Uppercase(A) :- 65 =< A, A =< 90.      % 65 = 'A' ..  90 = 'Z'

p_number(A) :- 48 =< A, A =< 57.       % 48 = '0' ..  57 = '9'

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%                         Right Hand Side                         %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

p_Rhs([61|S],ST,S_,RHS) :-			% 61 =
	! , elim_space(S,W),
	p_Expr(W,ST,S2,Exp),
	elim_space(S2,W2),
	p_Where(W2,Exp,ST,S_,RHS).

p_Where([A,B,C,D,E|S],CondRhs,ST,S_,[NewCondRhs,LocDefs]) :-
	[A,B,C,D,E] == "where" , ! ,
	elim_space(S,W),
	p_LocalDefs(W,CondRhs,ST,S_,NewCondRhs,LocDefs).

p_Where(W,Exp,_,W,[Exp,[]]).

p_Expr(W,ST,S_,Exp) :-			% "Expr2 '&' Exp | Exp2 "
	p_Expr2(W,ST,S,Exp1),
	elim_space(S,W2),		% verifica se expr appena riconosciuta 
	( [38|S2] = W2,			% e' seguita da '&'
	  ! , elim_space(S2,W3),
	  p_Expr(W3,ST,S_,Exp2),
	  Exp = apply(f("&"),[Exp1,Exp2])
	; Exp= Exp1, W2 = S_
	).

p_Expr2(W,_,S_,apply(c("success"),[]) ) :-	% "'success'"
	append("success",S_,W), ! .
p_Expr2(W,ST,S_,Exp) :-			% "Expr3 '=:=' Exp3 | Exp3 "
	p_Expr3(W,ST,S,Exp1),
	elim_space(S,W2),		% verifica se expr appena riconosciuta 
	( W2 = [61,58,61|S2],			% e' seguita da '=:='
	  ! , elim_space(S2,W3),
	  p_Expr3(W3,ST,S_,Exp2),
	  Exp = apply(f("=:="),[Exp1,Exp2])
	; Exp= Exp1, W2 = S_
	).

p_Expr3(W,ST,S_,Exp) :-			% "FunExpr ..."
	p_Expr4(W,ST,S,Exp1),
	elim_space(S,W2),		% verifica se expr appena riconosciuta 
	p_InOp_Expr(W2,Exp1,ST,S_,Exp).	% e' seguita da un op. infisso 


p_Expr4(W,ST,S_,Exp) :-			% "Expr3 ':' Exp3 | Exp3 "
	p_FunExpr(W,ST,S,Exp1),
	elim_space(S,W2),		% verifica se expr appena riconosciuta 
	( W2 = [58|S2],			% e' seguita da ':'
	  ! , elim_space(S2,W3),
	  p_Expr4(W3,ST,S_,Exp2),
	  Exp = apply(c("cons"),[Exp1,Exp2])
	; Exp= Exp1, W2 = S_
	).

p_InOp_Expr(W,Exp1,ST,S_,NewExp) :- %  "... InfixopId FunExpr"
	p_InfixOpId(W,S,Op),
	(	% '=' e '&' da soli non sono op. infissi
	  ( Op = "=" ; Op = "&") , 
	  ! , S_ = W, NewExp = Exp1
	; elim_space(S,W2),
	  p_Expr4(W2,ST,S_,Exp2),
	  NewExp =apply(f(Op),[Exp1,Exp2])
	) .

p_InOp_Expr(S,Exp,_,S,Exp).		% Non ci sono operatori infissi


p_FunExpr(W,ST,S_,Word):-		% "Bool"
	p_Key_Words(W,ST,S_,Word), ! .

p_FunExpr(W,_,_,_):- 		% parole riservate
	( append("then",_,W)
	; append("else",_,W)
	; append("where",_,W)
	; append("success",_,W)
	),
	! , fail.

p_FunExpr([A,B|S],ST,S_,Res) :-
		% "if FunExpr then Expr else Expr"     "if FunExpr then Expr"
	[A,B]=="if", 
	! , elim_space(S,W),
	p_Expr3(W,ST,S2,Cond),
	elim_space(S2,[C,D,E,F|S3]),
	[C,D,E,F]=="then", 
	! , elim_space(S3,W2),
	p_Expr(W2,ST,S4,ThenExp),
	elim_space(S4,W3),
	( append("else",S5,W3),
	  !, elim_space(S5,W4),
	  Res = apply(f("ifelse"),[Cond,ThenExp,ElseExp]),
	  p_Expr(W4,ST,S_,ElseExp)
	; Res = apply(f("if"),[Cond,ThenExp]),S_ = W3
	).

p_FunExpr(W,ST,S_,apply(c(Id),LBExp)):-	% "DataConstrId {BasicE}" 
	p_DataConstrId(W,S,Id),
	verify_DCID(Id,ST,_),
	! , elim_space(S,W2) ,
	p_ListBE(W2,[],ST,S_,LBExp).

p_FunExpr(W,ST,S_,apply(f(Id),LBExp)):-	% "FunctionName {BasicE}"
	p_FunName(W,S,Id),
	! , elim_space(S,W2) ,
	p_ListBE(W2,[],ST,S_,LBExp).

p_FunExpr(W,ST,S_,FExp) :-p_BasicE(W,ST,S_,FExp).	% "BasicE"

% p_BasicBE/4

p_BasicE(W,_,_,_):- 		% parole riservate
	( append("then",_,W)
	; append("else",_,W)
	; append("where",_,W)
	; append("success",_,W)
	),
	! , fail.

p_BasicE([40|S],_,S_, apply(c("()"),[])):- % "()", con spazi in mezzo
	elim_space(S,W) ,
	W=[41|S_], ! .

p_BasicE([40|S],ST,S_,Exp) :-		% "'(' Expr ... " 
	elim_space(S,W) ,
	p_Expr(W,ST,S2,El),
	elim_space(S2,W2) ,
	p_E_Tuple_Prec(W2,El,ST,S_,Exp).

p_BasicE([91|S],_,S_,apply(c("nil"),[])) :-	% "'['  ']'"
	elim_space(S,W2) ,
	W2=[93|S_], ! .

p_BasicE([91|S],ST,S_,List):-		% "'['Expr ..."
	elim_space(S,W) ,
	p_Expr(W,ST,S2,Head),
	elim_space(S2,W2),
	p_SeqExpr(W2,Head, [93],c("cons"),ST,S_,List).	% 93 = ]  

p_BasicE(W,_,S_,Num) :-		% "Int"
	p_Int(W,S_,Num).

p_BasicE(W,_,S_,v(Id)) :-		% "VaribleId"
	p_VarId(W,S_,Id).

p_BasicE(W,ST,S_, apply(c(Id),[])):-	% "DataConstrID "
	p_DataConstrId(W,S_,Id),
	verify_DCID(Id,ST,_), ! .

p_BasicE(W,_,S_, apply(f(Id),[])):-	% "FunctionName "
	p_FunId(W,S_,Id).

p_Key_Words([108,101,106|S],ST,S_,Exp) :-		% "'let'..."
	elim_space(S,W),
	p_LocalDefs(W,ST,ST2,S2),
	elim_space(S2,W2),
	W2=[105,110|S3],
	elim_space(S3,W3),
	p_Expr(W3,ST2,S_,Exp).

p_Key_Words(W,_,S_,apply(c("success"),[])) :-	% "success"
	append("success",S_,W).

p_Key_Words(W,_,S_,apply(c("true"),[])) :-	% Bool = true
	append("true",S_,W).

p_Key_Words(W,_,S_,apply(c("false"),[])) :-	% Bool = false
	append("false",S_,W).

p_ListBE(W,AccBE,ST,S_,LBE) :-		% " ... {BasicExpr}"
	p_BasicE(W,ST,S2,BE),		% Legge la BasicExpr e la accumula
	!,				% se p_BasicE e' ok elimina alternative di p_ListBE
	append(AccBE,[BE],AccBE2),
	elim_space(S2,W2),
	p_ListBE(W2,AccBE2,ST,S_,LBE).

p_ListBE(W,LBE,_,W,LBE).			% Le BasicExpr sono finite, perche' p_BasicE fallisce

p_E_Tuple_Prec([44|S],Exp1,ST,S_,apply(c("()"),[Exp1,Rest]) ) :- 
						% " ... ',' Expr ..." 2. elem. di tupla
	elim_space(S,W),
	p_Expr(W,ST,S2,Exp2),
	elim_space(S2,W2),
	p_SeqExpr(W2,Exp2,[41],c("()"),ST,S_,Rest).

p_E_Tuple_Prec([41|S],Exp,_,S,Exp).		% fine riconoscimento di "'(' Expr ')'"

p_SeqExpr([44|S],Exp1,End,HeapApp,ST,S_, apply(HeapApp,[Exp1,Exp])):- 
	elim_space(S,W),		% "... {',' Expr ..."  44 = ,  164 = 
	p_Expr(W,ST,S2,Exp2),
	elim_space(S2,W2),
	p_SeqExpr(W2,Exp2,End,HeapApp,ST,S_,Exp).

p_SeqExpr([93|S],Exp1,[93],c("cons"),_,S,
		apply(c("cons"),[Exp1,apply(c("nil"),[])]) ).
			% seq finisce con 93 = ] 
p_SeqExpr([41|S],Exp1,[41],c("()"),_,S,Exp1 ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%                       Local declarations                        %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

p_LocalDefs([123|S],CondRhs,ST,S_,NewCondRhs,NewLocDefs) :-
		% 123 = }
	elim_space(S,W),
	p_ValueDecl(W,CondRhs,ST,S2,CondRhs2,LocDef),
	elim_space(S2,W2),
	p_SeqLD(W2,CondRhs2,[LocDef],ST,S_,NewCondRhs,NewLocDefs).

p_SeqLD([59|S],CondRhs,AccLocDef,ST,S_,NewCondRhs,NewLocDefs):-
		% 59 = ;
	elim_space(S,W),
	p_ValueDecl(W,CondRhs,ST,S2,CondRhs2,LocDef),! ,
	elim_space(S2,W2),
	append(AccLocDef,[LocDef],AccLocDef2),
	p_SeqLD(W2,CondRhs2,AccLocDef2,ST,S_,NewCondRhs,NewLocDefs).

p_SeqLD([125|S_],CondRhs,AccLocDef,_,S_,CondRhs,AccLocDef). % 125 = }

p_ValueDecl(W,CondRhs,_,S_,CondRhs,apply(c("free"),LId)):-	
		% "VariableId { ',' VariableId } 'free'"
	p_VarId(W,S,Id), 
	elim_space(S,W2),
	p_ListVarDecl(W2,[v(Id)],S_,LId).

p_ValueDecl(W,CondRhs,ST,S_,NewCondRhs,(Patt,NewExp)):-	
		%  Pattern '=' Expr [ 'where' LocalDefs  ]
	retract(parsWh(0)),
	assertz(parsWh(1)),
	p_Patt(W,ST,_,S,Patt), 
	retract(parsWh(1)),
	assertz(parsWh(0)),

	extract_DataC(Patt,[],LDC),
	analyze_DataC(CondRhs,LDC,NewCondRhs),

	elim_space(S,[61|S2]),
	elim_space(S2,W2),
	p_Expr(W2,ST,S3,Exp),
	elim_space(S3,W3),
	p_Where(W3,Exp,ST,S_,NewExp).

p_ValueDecl(W,CondRhs,ST,S_,CondRhs,(Lhs,Rhs)):-	
	p_Equat(W,ST,S_,Lhs,Rhs).

p_ListVarDecl([44|S],AccId,S_,LId):-
	elim_space(S,W),
	p_VarId(W,S2,Id), ! , 
	elim_space(S2,W2),
	append([v(Id)],AccId,AccId2),
	p_ListVarDecl(W2,AccId2,S_,LId).

p_ListVarDecl([A,B,C,D|S],LId,S,LId):-
	[A,B,C,D]=="free".

extract_DataC(apply(c(Id),Args),LDC,NewLDC):-
	reserved_DataC(Id), ! ,
	extract_DataC_Args(Args,LDC,NewLDC).
extract_DataC(apply(c(Id),Args),LDC,NewLDC):-
	!,append([Id],LDC,AccLDC),
	extract_DataC_Args(Args,AccLDC,NewLDC).
extract_DataC(apply(_,Args),LDC,NewLDC):-
	extract_DataC_Args(Args,LDC,NewLDC).
extract_DataC(v(_),LDC,LDC).

extract_DataC_Args([],LDC,LDC).
extract_DataC_Args([Arg|Tail],LDC,NewLDC):-
	extract_DataC(Arg,LDC,AccLDC),
	extract_DataC_Args(Tail,AccLDC,NewLDC).

analyze_DataC(v(Id),_,v(Id)).
analyze_DataC(apply(f(Id),Args),LDC,apply(c(Id),NewArgs)):-
	member_DC(LDC,Id), ! ,
	analyze_DataC_Args(Args,LDC,NewArgs).
analyze_DataC(apply(F,Args),LDC,apply(F,NewArgs)):-
	analyze_DataC_Args(Args,LDC,NewArgs).

analyze_DataC_Args([],_,[]).
analyze_DataC_Args([Arg|Tail],LCD,[NewArg|NewTail]):-
	analyze_DataC(Arg,LCD,NewArg),
	analyze_DataC_Args(Tail,LCD,NewTail).

member_DC([Id|_],Id):- ! .
member_DC([_|LCD],Id) :- member_DC(LCD,Id).


reserved_DataC(Id):-
	(Id=="nil";Id=="cons";Id=="true";Id=="false";Id=="if";Id=="ifelse"
	;Id=="then";Id=="else";Id=="free";Id=="success").
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%                   Data declarations                             %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

p_TypeDecl(W,ST,ST_,S_,RR) :-
	search_eq(W,S,R),
	elim_space(S,W2),
	p_DataConstrId(W2,S2,Id),
	append("data ",R,R2),
	append(R2,Id,R3),
	append([Id],ST,NewST),
	search_DCID(S2,S3,R4),
	append(R3,R4,R5),
	elim_space(S3,W3),
	p_TD_Rest(W3,NewST,R5,ST_,S_,RR).

p_TD_Rest([124|S],ST,R,ST_,S_,RR) :-
	! , elim_space(S,W),
	p_DataConstrId(W,S2,Id),
	append(R,Id,R2),
	append([Id],ST,NewST),
	search_DCID(S2,W2,R3),
	append([32|R2],R3,R4),
	p_TD_Rest(W2,NewST,R4,ST_,S_,RR).
p_TD_Rest(S_,ST_,R,ST_,S_,R).

search_eq([61|S_],S_,[61,32]):-!.		% 61 = =
search_eq([A|S],S_,[A|R]) :- 	
	search_eq(S,S_,R).

search_DCID([124|S_],[124|S_],[124]):-!.	% 124 = |
search_DCID([10|S_],[10|S_],[10]):-!. 		% 10 = /n
search_DCID([A|S],S_,[A|R]) :- 	
	search_DCID(S,S_,R).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%                           Untranslate                           %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

untranslate_rule(apply(f(Id),Args),RHS) :-
	untranslate_fun(Id),
	untranslate_args(Args),
	untranslate_rhs(RHS).

untranslate_fun(Id):-
	p_FunId(Id,_,_),
	! , format(Id,[]).

untranslate_fun(Id):-
	p_InfixOpId(Id,_,_),
	write('('),
	format(Id,[]),
	write(')').

untranslate_rhs([Exp,[]]) :- 
	! , write(' = '),	
	untranslate_expr(Exp).

untranslate_rhs([Exp,LocDefs]) :- 
	! , write(' = '),	
	untranslate_expr(Exp),
	write(' where '),
	untranslate_First_LocDefs(LocDefs).

untranslate_First_LocDefs([ValDecl|LocDefs]):-
	write('{ '),
	untranslate_ValDecl(ValDecl),
	untranslate_LocDefs(LocDefs).

untranslate_LocDefs([]):-
	write(' } ').
untranslate_LocDefs([ValDecl|LocDefs]):-
	write(' ; '),
	untranslate_ValDecl(ValDecl),
	untranslate_LocDefs(LocDefs).

untranslate_expr(v(Var)) :- !,format(Var,[]).

untranslate_expr(apply(c("nil"),[])) :- !, write('[]').
untranslate_expr(apply(c("cons"),Args)) :- !, untranslate_list(Args,[]).
untranslate_expr(apply(c("true"),[]))  :- !, format("true",[]).
untranslate_expr(apply(c("false"),[])) :- !, format("false",[]).
untranslate_expr(apply(c("success"),[])) :- !, format("success",[]).
untranslate_expr(apply(c("()"),Args)) :-
	write('('),untranslate_explist(Args,c("()")),write(')').
untranslate_expr(apply(c([Head|Tail]),[])) :- 
	( p_number(Head),!
	; Head == 45 , [H:_]=Tail, p_number(H)
	),
	! , format([Head|Tail],[]).
untranslate_expr(apply(c(Name),[])) :- 
	format(Name,[]).
untranslate_expr(apply(c(Name),Args)) :- 
	write('('),format(Name,[]),untranslate_args(Args),write(')').
untranslate_expr(apply(f("&"),[Exp1,Exp2]) ) :-
	untranslate_expr(Exp1),
	write(' & '),
	untranslate_expr(Exp2).
untranslate_expr(apply(f("=:="),[Exp1,Exp2])) :- 
	untranslate_expr(Exp1),write(' =:= '),untranslate_expr(Exp2).
untranslate_expr(apply(f("ifelse"),[Cond,Exp1,Exp2]) ) :- 
	write('if '), untranslate_expr(Cond) ,
	write(' then '), untranslate_expr(Exp1) ,
	write(' else '), untranslate_expr(Exp2) .
untranslate_expr(apply(f("if"),[Cond,Exp]) ) :- 
	write('if '), untranslate_expr(Cond) ,
	write(' then '), untranslate_expr(Exp).

untranslate_expr(apply(f(Name),[])) :- 
	untranslate_fun(Name).
untranslate_expr(apply(f(Name),Args)) :- 
	write('('),untranslate_fun(Name),untranslate_args(Args),write(')').

untranslate_ValDecl(apply(c("free"),LId)):-
	untranslate_elements(LId), format(" free",[]).
untranslate_ValDecl((Lhs,Rhs)):-
	apply(f(_),_) = Lhs,
	! ,untranslate_rule(Lhs,Rhs).
untranslate_ValDecl((Patt,Exp)):-
	untranslate_expr(Patt),
	untranslate_rhs(Exp).

untranslate_args([]).
untranslate_args([A|RA]) :-
	write(' '), 
	untranslate_expr(A),
	untranslate_args(RA).
	
untranslate_elements([]).
untranslate_elements([A]) :-
	untranslate_expr(A).
untranslate_elements([A,B|RA]) :-
	untranslate_expr(A),
	write(','),
	untranslate_elements([B|RA]).
	
% untranslate_list([],_).
untranslate_list([Head,apply(c("cons"),Args)],LEl) :-!,
	append(LEl,[Head],AccLEl),
	untranslate_list(Args,AccLEl).
untranslate_list([Head,apply(c("nil"),[])],LEl) :- !,
	write('['),
	write_list(LEl,[44]),
	untranslate_expr(Head),
	write(']').
untranslate_list([Head,Tail],LEl) :-
	append(LEl,[Head],AccLEl),
	append(AccLEl,[Tail],AccLEl2),
	write_list(AccLEl2,[58]).
write_list([],_).
write_list([X],":"):-
	untranslate_expr(X).
write_list([Head|Tail],":"):-
	write('('),
	untranslate_expr(Head),
	format(":",[]),
	write_list(Tail,":"),
	write(')').
write_list([Head|Tail],Sep):-
	untranslate_expr(Head),
	format(Sep,[]),
	write_list(Tail,Sep).
untranslate_explist([],_).
untranslate_explist(Exp,DataC):- untranslate_explist2(Exp,DataC).

% untranslate_explist2([Head,[]],_) :- 
%	untranslate_expr(Head).
untranslate_explist2([Head,apply(Sym,Tail)],Sym) :- 
	untranslate_expr(Head),
	write(','),
	untranslate_explist2(Tail,Sym).
untranslate_explist2([Head,Tail],_) :- 
	untranslate_expr(Head),
	write(','),
	untranslate_expr(Tail).

%-- untranslate_dt => DEFTREES 
untranslate_dt( rule(NR,P,N) ) :- 
	untranslate_dt(rule(NR,P,N),0).
untranslate_dt( branch(P,O,Dts) ) :- untranslate_dt(branch(P,O,Dts),0).

untranslate_dt( rule(NR,apply(f(Id),Args),Rhs), Spaces ) :- 
	tab(Spaces),
        write('rule('), write(NR), write(','),	
	untranslate_fun(Id), untranslate_args(Args), 
	untranslate_rhs(Rhs), write(')'),nl.
untranslate_dt( branch(P,O,Dts), Spaces ) :- 
	tab(Spaces), Spaces1 is Spaces + 1,
        write('branch('), untranslate_expr(P), write(','),
        untranslate_Ox(O), write(',['),nl,
        untranslate_dtL( Dts, Spaces1 ),
	write('])').
                
untranslate_dtL( [], _ ).
untranslate_dtL( [X|Xs], Spaces ) :- 
	untranslate_dt(X, Spaces), untranslate_dtL(Xs,Spaces).

%-- untranslate_erL => STATI
untranslate_erL( []).
untranslate_erL( [X|Xs]) :- 
	untranslate_er(X), nl, untranslate_erL(Xs).

untranslate_er(er([],Rhs)):- 
	write('er( [] , '),untranslate_rhs(Rhs),write(' )').
untranslate_er(er(Subst,Rhs)):- 
	write('er( '), 
	untranslate_subL(Subst),write(', '),untranslate_rhs(Rhs),write(' )').

%-- untranslate_subL => SOSTITUZIONI
untranslate_subL( [Sub]):-
	! , untranslate_sub(Sub).
untranslate_subL([Sub|Subs]):-
	untranslate_sub(Sub),write(', '),untranslate_subL(Subs).

untranslate_sub((v(M),Exp)):-
	write('( '),untranslate_expr(v(M)),
	write('->'),untranslate_expr(Exp),write(' )').

% Ocurrence
untranslate_Ox( [] ).
untranslate_Ox( [X] ) :- write(X).
untranslate_Ox( [X,Y|Xs] ) :- write(X), write('.'), untranslate_Ox([Y|Xs]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%                              Misc                               %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

	
elim_space([32|S2],W) :-
	! , elim_space(S2,W).
elim_space(W,W).

verify_DCID(Id,ST,ST_):-
	parsWh(1), ! ,
	append([Id],ST,ST_).
verify_DCID(Id,ST,ST):-
	search_DC(ST,Id).

search_DC([Id|_],Id):-!.
search_DC([_|ST],Id) :-
	search_DC(ST,Id).
search_DC([],_):- fail.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%                              MENU                              %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

menu :- nl,nl,
  nl,nl,write('************************************************************'),
  nl,   write('*** load   --> Read R0 from a file                       ***'),
/*  nl,   write('***                                                      ***'),
  nl,   write('*** read   --> Read R0 from the standard input           ***'),
  nl,   write('***                                                      ***'),
  nl,   write('*** cond   --> Set the applicability conditions          ***'),
*/  
  nl,   write('*** clean  --> Clean the system                          ***'),
  nl,   write('*** intr   --> Definition introduction                   ***'),
  nl,   write('*** elim   --> Definition elimination                    ***'),
  nl,   write('*** list   --> List the rules of a program               ***'),
  nl,   write('*** dt     --> List a definitional tree                  ***'),
  nl,   write('*** unfold --> Lazy unfolding                            ***'),
  nl,   write('*** fold   --> Fold two rules                            ***'),
  nl,   write('*** acomp  --> Automatic Composition Strategiy           ***'),
  nl,   write('*** comp   --> Composition Strategiy                     ***'),
  nl,   write('*** tup    --> Tupling Strategiy                         ***'),
/*  nl,   write('***                                                      ***'),
  nl,   write('*** lastp  --> Last program of the sequence              ***'),
  nl,   write('***                                                      ***'),
  nl,   write('*** condq  --> Consult applicability conditions          ***'),
*/  
  nl,   write('*** auto   --> Strategy with/without message             ***'),
  nl,   write('*** exit   --> Leave the program                         ***'),
  nl,   write('************************************************************'),
  nl,nl,write('    OPTION: '),read(Option),option(Option),menu.

option(dt) :-
	nl, nl, write('    Operator: '),read(Name),
 	nl, write('    Program: '),read(Prog),

        name(Name,F),

	( find_dt(Prog,F,_,DT),
	  !, format("~s:\n",[F]), untranslate_dt(DT), nl
	; nl,nl,write('    ERROR: no definitional tree')
	),
	press.

option(load) :- 
	clean,
	nl, nl, write('    Filename: '),read(File),
	see(File),
	load_program([],Program),
	seen,
	elim_space(Program,ProgramW), 
	( [10|ProgramS]=ProgramW, 
	  ! ,elim_space(ProgramS,ProgramOk)
	; ProgramOk=ProgramW
	),
	parser(ProgramOk,1),
	! , 
	( []==ProgramOk,
	  nl,nl,write('    Empty program')
	; call_preprocess
	),
	press.

option(load) :- !,
	nl,nl,write('    ERROR: Parser error'),warning,
	clean,press.


option(read) :- 
	clean,
	nl,nl,write('Introduce program rules (quit to finish):'),nl,
	initial_program,!,
	call_preprocess,
	press.

option(read) :- !,
	nl,nl,write('    ERROR: Parser error'), warning,
	clean,press.


option(cond) :- !,
	nl,nl,write('    Applicability conditions? (yes/no): '), read(C),
	((C = 'yes'; C = 'y'),!,retractall(conditions),assert(conditions),nl,
	  write('    MESSAGE: Applicability conditions established');
	 (C = 'no'; C = 'n'),!,retractall(conditions),nl,
	  write('    MESSAGE: Without applicability conditions');
	  nl,write('    ERROR: Typed '),write(C)),
	nl,press.


option(clean) :- 
	!,clean,retractall(conditions),
	nl,nl,write('    MESSAGE: System cleaned'),
	press.

option(intr) :- !,nl,nl,  	  
  nl,nl,write('============================================================'),
  nl,   write('===               DEFINITION INTRODUCTION                ==='),
  nl,   write('============================================================'),
  nl,nl,write('        Definition of new functions (quit. to finish)'),nl,
  introduction(1,[]),
  nl,nl,write('============================================================'),
  nl,   write('===            END OF DEFINITION INTRODUCTION            ==='),
  nl,   write('============================================================'),
  press.

option(elim) :- !,nl,nl,
  nl,nl,write('============================================================'),
  nl,   write('===                DEFINITION ELIMINATION                ==='),
  nl,   write('============================================================'),
  elimination,
  nl,nl,write('============================================================'),
  nl,   write('===            END OF DEFINITION ELIMINATION             ==='),
  nl,   write('============================================================'),
  press.

option(list) :- !,nl,nl,
  nl,nl,write('============================================================'),
  nl,   write('===                  LIST OF THE RULES                   ==='),
  nl,   write('============================================================'),	
  show_rules,
  nl,nl,write('============================================================'),
  nl,   write('===                END OF LISTING RULES                  ==='),
  nl,   write('============================================================'),
  press.

option(unfold) :- !,nl,nl, 
  nl,nl,write('============================================================'),
  nl,   write('===                      UNFOLDING                       ==='),
  nl,   write('============================================================'),
  unfolding,
  nl,nl,write('============================================================'),
  nl,   write('===                   END OF UNFOLDING                   ==='),
  nl,   write('============================================================'),
  press.

option(fold) :- !,nl,nl, 
  nl,nl,write('============================================================'),
  nl,   write('===                       FOLDING                        ==='),
  nl,   write('============================================================'),
  folding,
  nl,nl,write('============================================================'),
  nl,   write('===                    END OF FOLDING                    ==='),
  nl,   write('============================================================'),
  press.

option(acomp) :- !,nl,nl, 
  nl,nl,write('============================================================'),
  nl,   write('===            AUTOMATIC COMPOSITION STRATEGY            ==='),
  nl,   write('============================================================'),
  nl,
  autocomposition,
  nl,nl,write('============================================================'),
  nl,   write('===         END OF AUTOMATIC COMPOSITION STRATEGY        ==='),
  nl,   write('============================================================'),
  press.

option(comp) :- !,nl,nl, 
  nl,nl,write('============================================================'),
  nl,   write('===                 COMPOSITION STRATEGY                 ==='),
  nl,   write('============================================================'),
  nl,
  composition,
  nl,nl,write('============================================================'),
  nl,   write('===             END OF COMPOSITION STRATEGY              ==='),
  nl,   write('============================================================'),
  press.

option(comp) :- !,nl,nl, 
  nl,nl,write('============================================================'),
  nl,   write('===                    TUPLING STRATEGY                   ==='),
  nl,   write('============================================================'),
  nl,
  tupling,
  nl,nl,write('============================================================'),
  nl,   write('===                END OF TUPLING STRATEGY               ==='),
  nl,   write('============================================================'),
  press.

option(auto) :- 
 	  nl,nl,write('    MESSAGE: Now transformation strategy is '),
	( auto(no), 
	  ! , write('with messages')
	; auto(yes),
 	  ! , write('without messages')
	),
	nl,write('             Change ? (yes/no): '), read(C),
	( ( C = 'yes'; C = 'y'),
	  !,
	  ( auto(yes), 
	    ! , retract(auto(yes)),
	    assertz(auto(no)),
 	    nl,nl,write('    MESSAGE: Transformation strategy with messages')
	  ; auto(no),
	    ! , retract(auto(no)),
	    assertz(auto(yes)),
 	    nl,nl,write('    MESSAGE: Transformation strategy without messages')
	  )
	; (C = 'no'; C = 'n'),!,true
	; nl,write('    ERROR: Typed '),write(C)
	),
	nl,press.

option(lastp) :- 
	last_program(LP),
	nl,nl,write('    LAST PROGRAM OF THE SEQUENCE: '),
	write(LP),press.

option(lastp) :- !,
	nl,nl,write('    MESSAGE: Transformation not started'),
	press.

option(condq) :- 
	(conditions,!,nl,nl,write('    Applicability conditions stablished');
	 !,nl,nl,write('    Without applicability conditions')),
	press.

option(exit) :- !,
    nl,nl,nl,write('If you want to come back to the menu, write "menu."'),
    nl,nl,nl,retractall(conditions),abort.

option(X) :- 
    nl,nl,write('    ERROR: Option '),write(X),write(' does not exist'),
    press.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%   CONDITIONS ON PROGRAMS (left-linearity and non overlapping)  %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

conditions(N,LHS) :-
	\+(left_linear(LHS)),!,
	nl,write('    ERROR: The lhs of the rule '),write(N),
	write(' is not left linear'),nl,
	fail.


conditions(N,LHS) :-
	overlapping(N,LHS),!,
	nl,write('    ERROR: The lhs of the rule '),write(N),
	write(' overlaps'),nl,
	fail.

conditions(_,_).

left_linear(LHS) :-
	extract_vars(LHS,VarsArgs),
	elim_rep(VarsArgs,VarsArgs).

overlapping(Rule,LHS) :- overlapping(1,Rule,LHS).

overlapping(N,_,LHS) :- 
	look_for_rule(_,N,Lhs_ed,_),
	( extract_occurs(LHS,OccurLHS),
	  unify_ov(Lhs_ed,OccurLHS),
	  !
	; extract_occurs(Lhs_ed,OccurLhs_ed),
	  unify_ov(LHS,OccurLhs_ed),
	  !
	).

overlapping(N,Rule,LHS) :-
	M is N+1, M < Rule,!,overlapping(M,Rule,LHS).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%               LOAD PROGRAM                                     %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

load_program(Aux,Program) :-
	get0(C),
	filter_ch(C,Aux,Program).

filter_ch(C,Aux,Program) :-
	( C > 31 , C <127, ! , append(Aux,[C],Result)
	; C == 10, ! ,append(Aux,[C],Aux2),filter_rt(Aux2,Result,Stop)
	),
	( Stop == "true", ! ,Program = Result
	; load_program(Result,Program)
	).

filter_ch(C,Program,Program) :-
	C== -1 , ! .

filter_ch(_,Aux,Program) :-
	load_program(Aux,Program).
filter_rt(Aux,Program,Stop):-
	get0(C),
	( C > 32 , C <127, !,append(Aux,[C],Program), Stop="false"
	; C == -1, ! , Program = Aux, Stop = "true"
	; filter_rt(Aux,Program,Stop)
	).

initial_program :- 
	read_rules([],Program),
	parser(Program,1).

read_rules(Prog,Program) :-
	nl,nl,write('"Rule". : '),read(Rule),
	(Rule='quit',!,Prog=Program;
	append(Prog,Rule,NewProg),read_rules(NewProg,Program)).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%           PRE-PROCESS                                          %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

call_preprocess:-
	assert(last_rule(0)),	  
	assert(last_program(-1)), 
	pre_process,
	deftrees,  
	! , nl,nl,write('    MESSAGE: Program loaded'),nl.

call_preprocess :-		
	clean,
	warning.

pre_process :- pre_process(1).

pre_process(N) :-
	rule(N,LHS,RHS),	

	( 
	 conditions,conditions(N,LHS),!;
	 \+(conditions),!
	),
	insert_new_rule(pair,N,LHS,RHS,_),
	M is N+1,
	pre_process(M).	

pre_process(N) :-
	rule(N,_,_),conditions,!,fail.

pre_process(_) :-
	!,
	retract(last_program(_)),
	assert(last_program(0)),
	last_rule(N),
	assert(last_rule_R0(N)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%               DEFINITION INTRODUCTION                          %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

introduction(N,Rules):-
	nl, nl, write('    New rule ("Rule".) : '), 
	read(NewRule),
	process_introd(NewRule,N,Rules).

process_introd(quit,_,Rules) :- 
	build_intr(Rules), ! .

	process_introd(quit,_,_) :- !,
	nl,nl,write('    MESSAGE: Any rule asserted').

process_introd(NewRule,N,Rules) :-
	parser(NewRule,N),!,
	rule(N,Lhs,Rhs),
	Rules2 = [rule(intr(N),Lhs,Rhs)|Rules],
	M is N + 1,
	introduction(M, Rules2).

process_introd(New_Rule,N,Rules) :-
	nl,nl,write('    ERROR: Parser error with the rule '),
	format("~s\n",New_Rule),
	introduction(N,Rules).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%               DEFINITION ELIMINATION                           %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

elimination :-
	nl,nl,write('    Introduce the symbol and its arity ("f/n"): '),
	read(String),
	string2term(String,Term),
	process_del(Term),!,
	retract(last_program(LP)),
	NP is LP+1,		
	assert(last_program(NP)).

elimination.

process_del(apply(f(X),Args)) :-
	is_in_R0(apply(f(X),Args)),!,
	nl,nl,write('    ERROR: Symbol '),
	format("~s",X),
	num_args(Args, Num_Args), write('/'),write(Num_Args),
	write(' is in the initial program'),fail.	

process_del(apply(f(X),Args)) :-
	rhs_BD(apply(f(X),Args)),!,
	nl,nl,write('    ERROR: Symbol '),
	format("~s",X),
	num_args(Args, Num_Args), write('/'),write(Num_Args),
	write(' is in the rhs of any rule'),fail.	

process_del(apply(f(F),Args)) :-
	num_args(Args,N),
	last_program(LP),
	find_dt(LP,F,N,DT),
	del_rl_dt(DT),
	NP is LP +1,
	assert(elim_dt(NP,F,N)),
	assertz(elim_rule),
	nl,write('    MESSAGE: Deleted rules'),nl.

process_del(apply(f(X),Args)) :-
	nl,nl,write('    ERROR: There is not any rule defining symbol '),
	format("~s",X),
	num_args(Args, Num_Args), write('/'),write(Num_Args),fail.	

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%                      RULE LISTING                              %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

show_rules :-
	last_rule(_),!,
	nl,nl,write('    Index of the program: '),read(Index),
	continue_show_rules(Index).

show_rules :- 
	nl,nl,write('    MESSAGE: There is not any rule').

continue_show_rules(Index) :-
	check(Index),!,
	nl,nl,write('    ERROR: R'),write(Index),
	write(' do not belong to the transformation sequence').

continue_show_rules(Index) :-
	nl,nl,write('NUMBER     RULE'),
	nl,show_rules(Index).

show_rules(K) :- show_rules(K,1).

show_rules(K,N) :-
	look_for_rule(P,N,_,_),
	P>K,!.

show_rules(_,N) :-
	last_rule(LR), N >LR,!.

show_rules(K,N) :-
	find_rule(K,N,LHS,RHS),!,
	( N<10,!,nl,write('   ')
	; nl,write('  ')
	),
	write(N),write('       '),
	untranslate_rule(LHS,RHS), nl,
	M is N+1, show_rules(K,M).

show_rules(K,N) :- 
	M is N+1, show_rules(K,M).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%                     UNFOLDING RULE                             %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

unfolding :- 
	nl,nl,write('    Rule Number: '),read(Rule),
	continue_unfolding(Rule).

continue_unfolding(Rule) :-
	last_program(LP),
	elimq(LP,Rule),!,
	nl,nl,write('    ERROR: Rule not found'),nl.

continue_unfolding(Rule) :-
	( last_rule(LR),Rule>LR, !
	; Rule = 0, !
	),
	nl,nl,write('    ERROR: Rule does not exist'),nl.

continue_unfolding(Rule) :- 
	look_for_rule(_,Rule,_,Rhs), is_hnf(Rhs),
	nl,nl,write('    ERROR: Unfolding rule is in hnf'),nl.


continue_unfolding(Rule):-
	last_program(Program),
	assert(selected_program(Program)),
	assert(selected_rule(Rule)),
	unfolding(Rule),
	! ,
	retract(selected_program(_)),
	retract(selected_rule(_)),
	show_and_save_u.

continue_unfolding(_):-
	nl,nl,write('    MESSAGE: Unfolding without effect').

unfolding(Rule) :-
        look_for_rule(_,Rule,_,Rhs),
        narrow_step_Rhs(Rhs,Result), 
        end_unfolding(Rule,Result). 
 
end_unfolding(Rule,[]):- 
        last_program(LP), NP is LP+1,    
        assertz(elim(NP,Rule)), 
        retract(last_program(LP)), 
        NP is LP+1, 
        assert(last_program(NP)). 
 
end_unfolding(Rule,Result):- 
    look_for_rule(_,Rule,LHS,_), 
    separa(Result,Derivs,Subs), 
    apply_subs(LHS,Subs,LHS_Subs), 
	merge(LHS_Subs,Derivs,unfold(Rule),Rules),
	build_unf(Rules),
	last_program(LP),  
	assertz(elim(LP,Rule)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%           FOLDING RULE                                         %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

folding :-
	elim_rule,
	!,
	nl,nl,write('    MESSAGE: Folding forbidden'),
	nl,nl,write('(folding after elimination without applicability conditions)').

folding :-
	nl,nl,write('    Folded rule number: '),read(FoldedRule),
	nl,   write('    Folding rule number: '),read(FoldingRule),
	folding(FoldedRule,FoldingRule),!,
	show_and_save_f.

folding.

folding(FoldedRule,FoldingRule) :- 
	last_rule(LR),
	(FoldedRule>LR,!,
	 nl,nl,write('    ERROR: Folded rule '),write(FoldedRule),
	 write(' out of bounds (last rule = '),write(LR),write(')');
	 FoldingRule>LR,!,
	 nl,nl,write('    ERROR: Folding rule '),write(FoldingRule),
	 write(' out of bounds (last rule = '),write(LR),write(')')
	),fail.

folding(FoldedRule,_) :-
	\+(find_rule_all(_,FoldedRule,_,_)),!,

	nl,nl,write('    ERROR: Rule '),write(FoldedRule),
	write(' does not exist'), fail.

/*folding(_,FoldingRule) :-

	\+((pair(P,rule(FoldingRule,_,_)),P>0)),!,
	
	nl,nl,
	write('    ERROR: Folding rule must be a definition rule'),fail.
*/

folding(FoldedRule,_) :- 
	self_folding(FoldedRule),!,
	nl,nl,write('    ERROR: Danger of self-folding'),fail.

folding(FoldedRule,FoldingRule) :-
	look_for_rule(_,FoldedRule,Lhs_ed,Rhs_ed),
	look_for_rule(_,FoldingRule,TLhs_ing,TRhs_ing),
	last_prg_vble(M),
	stand_apart(M,TRhs_ing,Rhs_ing),
	stand_apart(M,TLhs_ing,Lhs_ing),
	extract_positions(Rhs_ed,Positions),
	cont_try_f(Positions,1,Rhs_ed,Lhs_ing,Rhs_ing,_,Rhs_folded),
	\+(Rhs_folded = Rhs_ed), \+(Rhs_folded = inf),
	last_program(LP),
	apply(f(F),Args) = Lhs_ed,
	num_args(Args,NArgs),
	find_dt(LP,F,NArgs,A),
	update_prg_vbles,
	search_rl_dt(A,FoldedRule,[rule(fold(FoldedRule),Lhs_ed,Rhs_folded)],T),
	! ,
	NP is LP + 1,
	assertz(dt(NP,F,NArgs,T)),
	retract(last_program(LP)),
	assert(last_program(NP)),
	assertz(elim(NP,FoldedRule)).

folding(FoldedRule,FoldingRule) :-
	nl,nl,write('    ERROR: Impossible folding between rules '),
	write(FoldedRule),write(' and '),write(FoldingRule),fail.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%                      COMPOSITION STRATEGY                      %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

autocomposition :-
	last_rule(LR),
	LR==0,
	nl,nl,write('    ERROR: No Rule present '),nl, ! , fail.
autocomposition :-
	last_rule(LR), Max is LR +1 ,
	autocomp(1,Max).

autocomp(Max,Max):-
	( auto(no),
	  ! , nl,nl,write('    ERROR: No composition possible '),nl
	; auto(yes)
	).
autocomp(N,_):-
	last_program(LP),
	find_rule(LP,N,_,[Rhs,LD]),
	double_call(Rhs,N,LD),
	! .
autocomp(N,Max):-
	M is N + 1,
	autocomp(M,Max).

double_call(apply(f(F1),Args),N,LD):-
	( no_predef(F1),
	  double_call2(Args,N,apply(f(F1),Args),LD),
	  !
	; double_calls(Args,N,LD)
	).
double_call(apply(c(_),Args),N,LD):-
	double_calls(Args,N,LD).

double_calls([],_,_):- fail.
double_calls([A|Args],N,LD):-
	( double_call(A,N,LD),
	  ! 
	; double_calls(Args,N,LD)
	).

double_call2([],_,_,_):-fail.
double_call2([Arg|Other],N,Exp,LD):-
	last_prg_vble(M),
	( apply(f(F2),_) = Arg,
	  no_predef(F2), 
	  stand_apart(M,Exp,ExpSt),
	  Rhs_ing = [ExpSt,[]] ,
	  extract_vars(ExpSt,Vars),
	  elim_rep(Vars,Vars2),
	  elim_v_anony(Vars2,Vars3),
	  elim_d_free(LD,Vars3,Vars4),
	  append("new",M,F),
	  update_prg_vbles,
	  Lhs_ing = apply(f(F),Vars4),
	  assert(rule(1,Lhs_ing,Rhs_ing)),
	  comp_rls(N,Lhs_ing,Rhs_ing),
	  ! 
	; retract(last_prg_vble(_)),
	  assertz(last_prg_vble(M)),
	  double_call2(Other,N,Exp,LD)
	).

composition :-
	last_rule(LR),
	LR==0,
	nl,nl,write('    ERROR: No Rule present '),nl, ! , fail.

composition :-
	nl, nl, write('    New rule ("Rule".) : '), 
	read(NewRule),
	( parser_ok(NewRule), 
	  rule(1,Lhs_ing,Rhs_ing),
	  %nl, nl, write('    Rule Number : '), 
	  %read(NR),
	  %comp_rls(NR,Lhs_ing,Rhs_ing),  
	  comp_rls(1,Lhs_ing,Rhs_ing),
	  ! 
	; ( auto(no),
	    ! , nl,nl,write('    ERROR: No composition possible '),nl
	  ; auto(yes)
	  )
	).


comp_rls(NR,Lhs_ing,Rhs_ing):-
	( build_intr([rule(intr(1),Lhs_ing,Rhs_ing)]),
	  !
	; retractall(rule(_,_,_)),
	  !, fail
	),
	last_program(NP2),
	NP1 is NP2 -1,
	assert(selected_program(NP2)),
	apply(f(F_ing),Args)=Lhs_ing,
	num_args(Args,N_ing),
	pair(NP2,rule(NR_ing,apply(f(F_ing),_),_)),
	ctrl_folded(F_ing,N_ing,NP1,NR,Lhs_ed,Rhs_ed),
	try_folding(Rhs_ed,Lhs_ing,Rhs_ing,NR,Rhs_ed_folded),
	repeat_unf([er([],Rhs_ing)],0,[],NR_ing,States,Stopped,Sts_folded),
	( 
	  [er(_,_)|_]=Sts_folded,
	  ! , 
	  NP3 is NP2+1,
	  assertz(elim(NP3,NR_ing)),
	  assertz(elim_dt(NP3,F_ing,N_ing)),
	  append(States,Stopped,St),
	  append(St,Sts_folded,St2),
	  merge_er(St2,Lhs_ing,NR_ing,Rules),
	  build_unf(Rules),
	  retract(selected_program(_)),
	  change_type_rule(Sts_folded,NP3,NR_ing),
	  assertz(elim(NP3,NR)),
	  apply(f(F_ed),Args_ed) = Lhs_ed,
	  num_args(Args_ed,NArgs_ed),
	  find_dt(NP2,F_ed,NArgs_ed,DT),
	  retract(last_program(_)),
	  assert(last_program(NP2)),
	  search_rl_dt(DT,NR,[rule(fold(NR),Lhs_ed,Rhs_ed_folded)],NDT),
	  retract(last_program(_)),
	  assert(last_program(NP3)),
	  assertz(dt(NP3,F_ed,NArgs_ed,NDT)),
	  show_c
	; inf=Sts_folded,
	  ! ,
	  ( auto(no),
	    ! , nl,nl,write('    ERROR: Unfolding reach deep('),
	    deep(K), write(K), write(')'), nl
	  ; auto(yes)
	  ),
	  stop_comp(F_ing,N_ing,NP1),
	  fail

	; ( auto(no),
	    ! , nl,nl,write('    ERROR: Composition not possible at rule '),
	    write(NR), nl
	  ; auto(yes)
	  ),
	  stop_comp(F_ing,N_ing,NP1),
	  fail
	).

repeat_unf(_,N,_,_,[],[],inf):-
	deep(K), N>=K, ! .

repeat_unf(States,N,Stopped,NR_ing,StRes,StoppedF,Sts_folded):-
    narrow_step_State(States,Result,Stopped2),
	append(Stopped,Stopped2,Stopped3),
	( [] = Result, 		% no narrowing steps
	  ! , StoppedF=Stopped3, Sts_folded=[], StRes=[]
	; ( 
	    try_f_sts(Result,NR_ing,[],StRes,Sts_folded),
	    ! , StoppedF=Stopped3
	  ; 
	    M is N +1,
	    repeat_unf(Result,M,Stopped3,NR_ing,StRes,StoppedF,Sts_folded)
	  )
	).

narrow_step_State([],[],[]):- ! . 
narrow_step_State([er(Subst,Rhs)|State],Result4,Stopped):-
	\+(is_hnf(Rhs)),
	! , narrow_step_Rhs(Rhs,Result),
	uniun_SUBS_stL(Result,Subst,Result2),
	append(Result2,Result3,Result4),
	narrow_step_State(State,Result3,Stopped).
narrow_step_State([er(Subst,Rhs)|States],Result,[er(Subst,Rhs)|Stopped]):-
	% Rhs e' in hnf, inserisco lo stato nella lista dei bloccati 
	%  e passo al prossimo
	narrow_step_State(States,Result,Stopped). 

try_f_sts([],_,[],_,_):- !, fail.
try_f_sts([],_,Sts_folded,[],Sts_folded):- ! .

try_f_sts([er(Subst,Rhs_ed)|States],NR_ing,AccSts_folded,Sts_out,Sts_folded):-
	pair(_,rule(NR_ing,TLhs_ing,TRhs_ing)),
	last_prg_vble(M),
	stand_apart(M, TLhs_ing, Lhs_ing ),
	stand_apart(M, TRhs_ing, Rhs_ing ),
	extract_positions(Rhs_ed,Positions),
	cont_try_f(Positions,1,Rhs_ed,Lhs_ing,Rhs_ing,Theta,Rhs_folded),
	\+(Rhs_folded = Rhs_ed), 
	\+(Rhs_folded = inf),
	update_prg_vbles,
	substitute_SUBS(Theta,Subst,SubstTh),
	! ,
	AccSts_folded2 = [er(SubstTh,Rhs_folded)|AccSts_folded], 
	try_f_sts(States,NR_ing,AccSts_folded2,Sts_out,Sts_folded).

try_f_sts([er(Subst,Rhs)|States],NR_ing,AccSts_folded,
		[er(Subst,Rhs)|Sts_out],Sts_folded):-
	try_f_sts(States,NR_ing,AccSts_folded,Sts_out,Sts_folded).

cont_try_f(_,N,_,_,_,_,inf) :- deep(K), N>K, !.
cont_try_f([],_,R,_,_,[],R) :- ! .
cont_try_f([Pos|RestPos],N,Rhs_ed,Lhs_ing,Rhs_ing,Theta3,Rhs_folded2) :-
	subexpression(Rhs_ed,Pos,R_p),
	\+(is_hnf(R_p)),
	pattern_match_rhs(Rhs_ing,R_p,Theta),
	! ,
	substitute(Theta,Lhs_ing,Lhs_Subs),
	substitute_expr(Rhs_ed,Pos,Lhs_Subs,Rhs_folded),
	elim_pos(RestPos,Pos,RestPos2),
	extract_positions([Lhs_Subs,[]],Positions),
	append_pos(Positions,Pos,Positions2),
	append(Positions2,RestPos2,RestPos3),
	M is N+1 ,
	cont_try_f(RestPos3,M,Rhs_folded,Lhs_ing,Rhs_ing,Theta2,Rhs_folded2),
	substitute_SUBS(Theta2,Theta,Theta3),
	! .

cont_try_f([_|RestPos],N,Rhs_ed,Lhs_ing,Rhs_ing,Theta,Rhs_folded) :-
	cont_try_f(RestPos,N,Rhs_ed,Lhs_ing,Rhs_ing,Theta,Rhs_folded).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%                      AUXILIAR PROCEDURES                       %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

press :- nl,nl,nl,write('    Press Enter to continue ... '),get0(_).
warning :- 
	nl,nl,write('    WARNING: Do not start the transformation process').

zip_occ([],[],[]).
zip_occ([],[_|_],[]):- fail.
zip_occ([_|_],[],[]):- fail.
zip_occ([A|As],[B|Bs],[(A,B)|Rest]) :-
	zip_occ(As,Bs,Rest).

extract_vars(v(Var),[v(Var)]).

extract_vars(apply(_,Args),Vars) :-
	extract_vars2(Args,Vars).

extract_vars2([],[]).

extract_vars2([Arg|RestArgs],Vars) :-
	extract_vars(Arg,V_Arg),
	extract_vars2(RestArgs,V_RestArgs),
	append(V_Arg,V_RestArgs,Vars).

elim_rep(L1,L2) :- elim_rep(L1,[],L2).

elim_rep([],L,L).

elim_rep([H|T],Aux,L) :-
	member(H,Aux),!,
	elim_rep(T,Aux,L).

elim_rep([H|T],Aux,L) :- 
	append(Aux,[H],Aux2),
	elim_rep(T,Aux2,L).

extract_occurs(v(_),[]).

extract_occurs(apply(F,Args),[apply(F,Args)| RestOccurs]) :-
	extract_occur2(Args,RestOccurs).

extract_occur2([],[]).

extract_occur2([v(_) | RestArgs], Occurs) :- 
	extract_occur2(RestArgs,Occurs).

extract_occur2([apply(F,Args) | RestArgs], [apply(F,Args) | RestOccurs]) :-
	extract_occur2(Args,OccArgs),
	extract_occur2(RestArgs,OccRestArgs),
	append(OccArgs,OccRestArgs,RestOccurs).

unify_ov(Term,[Elem|_]) :-
	unification(Term,Elem),!.

unify_ov(Term,[_| RestElems]) :-
	unify(Term,RestElems).

unification(v(_),_) :- !.
unification(_,v(_)) :- !.
unification(apply(F,Args1),apply(F,Args2)) :-
	unification_args(Args1,Args2).

unification_args([],[]).
unification_args([Args1|RestArgs1],[Args2|RestArgs2]) :-
	unification(Args1,Args2),
	unification_args(RestArgs1,RestArgs2).

clean :-
	retractall(rule(_,_,_)),
	retractall(pair(_,_)),
	retractall(elim(_,_)),
	retractall(unf(_,_,_)),
	retractall(fold(_,_,_)),
	retractall(last_rule(_)),
	retractall(last_rule_R0(_)),
	retractall(last_program(_)),
	retractall(last_prg_vble(_)),
	retractall(elim_rule),
	retractall(selected_program(_)),
	retractall(dt(_,_,_,_)),
	retractall(elim_dt(_,_,_,_)),
	retractall(deep(_)),
	retractall(nInPos(_)),
	retractall(auto(_)),
	retractall(data_dec(_)),
	retractall(symtab(_)),
	assert(last_prg_vble("A")),
	assertz(nInPos(0)),
	assertz(auto(no)),
	assertz(deep(10)),
	assertz(data_dec([])),
	assertz(sym_tab([])).

analyze([],[]).

analyze([rule(intr(N),apply(f(X),Args),_)|Rules],NewRules) :-
	sequence_before(apply(f(X),Args)),!,
	retract(rule(N,_,_)),
	nl, write('    ERROR: '),
	format("~s",X),
	num_args(Args, Num_Args), write('/'),write(Num_Args),
	write(' is not a new function symbol'),nl,
	analyze(Rules,NewRules).

analyze([rule(intr(N),_,Rhs)|Rules],NewRules) :-
	\+(comprobar_rhs(Rhs)),!,
	retract(rule(N,_,_)),
	nl, write('    ERROR: Any symbol of the body is new'),nl,
	analyze(Rules,NewRules).

analyze([rule(intr(N),Lhs,_)|Rules],NewRules) :-
	conditions,\+(conditions(N,Lhs)),!,
	retract(rule(N,_,_)),
	analyze(Rules,NewRules).

analyze([Rule|Rules],[Rule|NewRules]) :-
	analyze(Rules,NewRules).

insert_new_rule(pair,N, LHS, RHS, NR) :-
	last_program(P), NP is P+1,
	retract(last_rule(R)), NR is R+1,
	assertz(last_rule(NR)),
	assertz(pair(NP,rule(NR,LHS,RHS))),
	retract(rule(N,LHS,RHS)).

insert_new_rule(unf, Father, LHS, RHS, NR) :-
	last_program(P), NP is P+1,
	retract(last_rule(R)), NR is R+1,
	assertz(last_rule(NR)),
	assertz(unf(NP,Father,rule(NR,LHS,RHS))).

insert_new_rule(fold, Father, LHS, RHS, NR) :-
	last_program(P), NP is P+1,
	retract(last_rule(R)), NR is R+1,
	assertz(last_rule(NR)),
	assertz(fold(NP,Father,rule(NR,LHS,RHS))).


comprobar_rhs([Cond,LocDefs]):-
	comprobar_body(Cond),
	comprobar_body2(LocDefs).

comprobar_body(v(_)).
comprobar_body(apply(f(X),Args)) :-
	! ,
	sequence_before(apply(f(X),Args)),!,
	comprobar_body2(Args).
comprobar_body((_,Rhs)) :-
	! ,
	comprobar_rhs(Rhs).
comprobar_body(apply(_,Args)) :- 	
	comprobar_body2(Args).

comprobar_body2([]) :- !.
comprobar_body2([Arg|Args]) :-
	comprobar_body(Arg),
	comprobar_body2(Args).

sequence_before(Symbol) :- sequence_before(Symbol,1).

sequence_before(Symbol,N) :-
	look_for_rule(_,N,Head,_),
	\+(elim(_,N)),
	same(Head,Symbol),
	! .

sequence_before(Symbol, N) :-
	M is N+1,
	last_rule(LR), M=<LR,
	! , sequence_before(Symbol,M).

same(apply(X,ArgsF),apply(X,ArgsG)) :-
	same_number_args(ArgsF,ArgsG).

same_number_args([],[]):-!.
same_number_args([_|RestF],[_|RestG]) :- same_number_args(RestF,RestG).


num_args(Args,N) :- num_args(Args,0,N).

num_args([],N,N) :- !.
num_args([_|Rest], N, Total) :- 
	M is N+1, 
	num_args(Rest, M, Total).


construir_termino(Symbol,Arity,apply(f(Symbol),Args)) :-
	num_args(Args,Arity).

is_in_R0(Term) :- is_in_R0(Term,1).

is_in_R0(Term,N) :-
	pair(0,rule(N,LHS,_)),
	same(Term,LHS),!.

is_in_R0(Term,N) :-
	pair(0,rule(N,_,RHS)),
	% verifico se Term compare nella rhs
	comprobar_term_rhs(Term,RHS),!.

is_in_R0(Term,N) :-	
	M is N+1,
	last_rule_R0(Last_R0), 	M =< Last_R0, !,
	is_in_R0(Term,M).

rhs_BD(Term) :- 
	last_rule_R0(R0),
	Start is R0+1,
	rhs_BD(Term,Start).

rhs_BD(Term,N) :-
	find_rule_all(P,N,LHS,RHS), P>0,
	comprobar_term_rhs(Term,RHS),
	\+(same(Term,LHS)),!. 
	
rhs_BD(Term,N) :-
	M is N+1, last_rule(LR), 
	M =< LR, !,
	rhs_BD(Term,M).

comprobar_term_rhs(Term,[Cond,LocDefs]) :- 
	( comprobar_body(Term,Cond), !
	; comprobar_body2(Term,LocDefs)
	).

comprobar_body(Term,Body) :- same(Term,Body),!.
comprobar_body(Term,(_,Rhs)):-
	! , comprobar_term_rhs(Term,Rhs).
comprobar_body(Term,apply(_,Args)):-
	comprobar_body2(Term,Args).

comprobar_body2(_,[]):-fail. 
comprobar_body2(Term,[Arg|Args]) :- 
	comprobar_body(Term,Arg),
	comprobar_body2(Term,Args).

find_rule_all(P,N,LHS,RHS) :-
	look_for_rule(P,N,LHS,RHS),
	\+(elim(_,N)).

find_rule(K,N,LHS,RHS) :-
	look_for_rule(_,N,LHS,RHS),
	\+(elimq(K,N)).

elimq(K,N) :- elim(P,N),P=<K.

look_for_rule(P,N,LHS,RHS) :-
	(
	 pair(P,rule(N,LHS,RHS)),!;
          unf(P,_,rule(N,LHS,RHS)),!;
	 fold(P,_,rule(N,LHS,RHS))
	).

check(Program) :- last_program(LP),Program>LP.

string2term(Rule,NewRule) :-
	obtain(Rule,Symbol,Arity),
	construir_termino(Symbol,Arity,NewRule).

obtain([S|Rest],[S|RestSymbol],Arity) :-
	S \== 47,!,
	obtain(Rest,RestSymbol,Arity).

obtain([47|Rest],[],Arity) :- 		% 47 = /
	obtain_arity(Rest,Arity).


obtain_arity(Rest,Arity) :- obtain_arity(Rest,0,Arity).
obtain_arity([],Arity,Arity) :- !.
obtain_arity([Num|Rest],Cont,Arity) :-
	NewCont is (Cont*10)+(Num-48),
	obtain_arity(Rest,NewCont,Arity).

separa([],[],[]).
separa([er(S,D)|Rest],[D|RestD],[S|RestS]) :-
	separa(Rest,RestD,RestS).

apply_subs(_,[],[]).
apply_subs(Exp,[Subs|Rest],[Subs_Exp|Rest_Subs_Exp]) :-
	substitute(Subs,Exp,Subs_Exp),
	apply_subs(Exp,Rest,Rest_Subs_Exp).

merge([],[],_,[]).
merge([H1|T1],[H2|T2],Rule,[rule(Rule,H1,H2)|Tail]):-
	merge(T1,T2,Rule,Tail).

is_hnf([Cond,LocDefs]):-
	is_hnf(Cond),
	is_hnf_arg(LocDefs).

is_hnf(apply(f("=:="),[apply(c(X),_),apply(c(X),_)])):- ! ,fail .
is_hnf(apply(f("=:="),[v(_),_])):- ! ,fail .
is_hnf(apply(f("=:="),[_,v(_)])):- ! ,fail .
is_hnf(apply(f("=:="),[Exp1,Exp2])):-
	! ,is_hnf(Exp1),
	is_hnf(Exp2).

is_hnf(apply(f("&"),[apply(c("success"),[]),_])):- ! , fail.
is_hnf(apply(f("&"),[v(_),_])):- ! , fail.
is_hnf(apply(f("&"),[_,apply(c("success"),[])])):- ! , fail.
is_hnf(apply(f("&"),[_,v(_)])):- ! , fail.
is_hnf(apply(f("&"),[Exp1,Exp2])):-
	! ,is_hnf(Exp1),
	is_hnf(Exp2).

is_hnf(apply(f("ifelse"),[apply(c(Cond),_),_,_])):- 
	( Cond="true" , !
	; Cond="false", !
	) , fail.
is_hnf(apply(f("ifelse"),[Exp,_,_])):-
	! , is_hnf(Exp).
is_hnf(apply(f("if"),[apply(c(Cond),_),_])):- 
	( Cond="true" , !
	; Cond="false", !
	) , fail.
is_hnf(apply(f("if"),[Exp,_])):-
	! , is_hnf(Exp).

is_hnf((_,Rhs)):-
	! , is_hnf(Rhs).

is_hnf(apply(f(_),_)):- fail.
is_hnf(v(_)).
is_hnf(apply(c(_),_)).

is_hnf_arg([]).
is_hnf_arg([H|T]):-
	is_hnf(H),
	is_hnf_arg(T).

show_and_save_u :- 
  nl, nl,
  nl,nl,write('************************************************************'),
  nl,   write('***                  RESULTS OF UNFOLDING                ***'),
  nl,   write('************************************************************'),
  nl,   write('***                                                      ***'),
  nl,   write('*** all   --> Show the unfolded program                  ***'),
  nl,   write('***                                                      ***'),
  nl,   write('*** unf   --> Show the unfolded rules                    ***'),
  nl,   write('***                                                      ***'),
  nl,   write('*** svall --> Save the unfolded program                  ***'),
  nl,   write('***                                                      ***'),
  nl,   write('*** svunf --> Save the unfolded rules                    ***'),
  nl,   write('***                                                      ***'),
  nl,   write('*** ret   --> Return to the main menu                    ***'),
  nl,   write('***                                                      ***'),
  nl,   write('************************************************************'),
  nl,   write('************************************************************'),
  nl,nl,write('    OPTION: '),read(Option),opt_u(Option).


opt_u(all) :- !,
	last_program(LP),nl,nl,write('NUMBER     RULE'),nl, 
	show_rules(LP),
	press,
	show_and_save_u.

opt_u(unf) :-
      ( 
	last_program(LP),
	unf(LP,_,rule(N,_,_)),!,
	nl,nl,write('NUMBER     RULE'),nl,
	show_rules(LP,N),
	press,
	show_and_save_u
      ;
	!,nl,nl,write('    MESSAGE: Only deletion of the unfolded rule'),
	press,
	show_and_save_u
      ).
opt_u(svall) :- !,
	nl,write('    Filename: '), read(File),
	tell(File),
	save_all,
	told,
	nl,write('    MESSAGE: File saved'),nl,
	press,
	show_and_save_u.
opt_u(svunf) :- !,
	nl,write('    Filename: '), read(File),
	tell(File),
	save_unf,
	told,
	nl,write('    MESSAGE: File saved'),nl,
	press,
	show_and_save_u.
opt_u(ret) :- !.
opt_u(X) :- 
	nl,nl,write('    ERROR: Option '),write(X),write(' does not exist'),
	press,
	show_and_save_u.

save_unf :- 
	last_program(LP),
	unf(LP,_,rule(N,_,_)),!,
	save_rules(LP,N).
save_unf.

save_all :- last_program(LP),
	data_dec(DD),
	format("~s\n",DD),
	save_rules(LP,1).


save_rules(_,N) :- last_rule(LR), N>LR,!.
save_rules(P,N) :-
	find_rule(P,N,LHS,RHS),!,
	untranslate_rule(LHS,RHS) , nl,
	M is N+1,save_rules(P,M).
save_rules(P,N) :-			% regola N eliminata
	M is N+1, save_rules(P,M).

self_folding(N) :-
	look_for_rule(_,N,LHS,_),
	\+(is_in_R0(LHS)),
	\+(at_least_one(N)).


at_least_one(N) :- unf(_,_,rule(N,_,_)),!.
at_least_one(N) :- fold(_,Father,rule(N,_,_)),at_least_one(Father).

extract_positions([Cond,LocDefs],Positions):-
	extract_positions([Cond],1,[],Positions1),
	extract_positions(LocDefs,1,[2],Positions2),
	append(Positions1,Positions2,Positions).

extract_positions([],_,_,[]) :- !.

extract_positions([v(_) | Rest],N,Pred,RestPos) :-
	! ,
	% estrae le pos. dai suoi fratelli
	M is N+1,
	extract_positions(Rest,M,Pred,RestPos).

extract_positions([(_,[Cond,LocDefs]) | Rest],N,Pred,Positions) :-
	! , 
	M is N+1,
	extract_positions(Rest,M,Pred,RestAux),
	append(Pred,[N,2],Pos1), 
	append(Pred,[N,2,2],Pos2),
	extract_positions([Cond],1,Pos1,RestAux2),
	extract_positions(LocDefs,1,Pos2,RestAux3),
	append(RestAux,RestAux2,AccPos),
	append(AccPos,RestAux3,Positions).

extract_positions([apply(c("free"),_) | Rest],N,Pred,RestPos) :-
	! ,
	M is N+1,
	extract_positions(Rest,M,Pred,RestPos).

extract_positions([apply(_,Args) | Rest],N,Pred,[Pos|Positions]) :-
	! ,
	append(Pred,[N],Pos),
	M is N+1,
	extract_positions(Rest,M,Pred,RestAux),
	append(RestAux,RestAux2,Positions),
	extract_positions(Args,1,Pos,RestAux2).

show_and_save_f :- 
  nl, nl,
  nl,nl,write('************************************************************'),
  nl,   write('***                   RESULTS OF FOLDING                 ***'),
  nl,   write('************************************************************'),
  nl,   write('***                                                      ***'),
  nl,   write('*** all    --> Show the folded program                   ***'),
  nl,   write('***                                                      ***'),
  nl,   write('*** fold   --> Show the folded rules                     ***'),
  nl,   write('***                                                      ***'),
  nl,   write('*** svall  --> Save the folded program                   ***'),
  nl,   write('***                                                      ***'),
  nl,   write('*** svfold --> Save the folded rules                     ***'),
  nl,   write('***                                                      ***'),
  nl,   write('*** ret    --> Return to the main menu                   ***'),
  nl,   write('***                                                      ***'),
  nl,   write('************************************************************'),
  nl,   write('************************************************************'),
  nl,nl,write('    OPTION: '),read(Option),opt_f(Option).

opt_f(all) :- !,
	last_program(LP),nl,nl,write('NUMBER     RULE'),nl, 
	show_rules(LP),
	press,
	show_and_save_f.
opt_f(fold) :-
      (
	last_program(LP),
	fold(LP,_,rule(N,_,_)),!,
	nl,nl,write('NUMBER     RULE'),nl,
	show_rules(LP,N),
	press,
	show_and_save_f
      ;
	!,nl,nl,write('    MESSAGE: Only deletion of the folded rule'),
	press,
	show_and_save_f
      ).
opt_f(svall) :- !,
	nl,write('    Filename: '), read(File),
	tell(File),
	save_all,
	told,
	nl,write('    MESSAGE: File saved'),nl,
	press,
	show_and_save_f.
opt_f(svfold) :- !,
	nl,write('    Filename: '), read(File),
	tell(File),
	save_fold,
	told,
	nl,write('    MESSAGE: File saved'),nl,
	press,
	show_and_save_f.
opt_f(ret) :- !.
opt_f(X) :- 
	nl,nl,write('    ERROR: Option '),write(X),write(' does not exist'),
	press,
	show_and_save_f.

save_fold :-
	last_program(LP),
	fold(LP,_,rule(N,_,_)),!,
	save_rules(LP,N),
save_fold.

no_predef(F):-
	F\=="&",F\=="==",F\=="+",F\=="-",F\=="*",F\=="div",F\=="%",F\==">",
	F\=="<",F\==">=",F\=="=<",F\=="=:=",F\=="ifelse",F\=="if".

elim_v_anony([],[]).
elim_v_anony([v(N)|Vars],VarsF):-
	( append("_",_,N),
	  ! , elim_v_anony(Vars,VarsF)
	; VarsF = [v(N)|Vars2],
	  elim_v_anony(Vars,Vars2)
	).	

elim_d_free([],Vars,Vars).
elim_d_free([LD|LDs],Vars,VarsF):-
	( apply(c("free"),VFs) = LD ,
	  !, extr_list(Vars,VFs,Vars2)
	; Vars2 = Vars
	) ,
	elim_d_free(LDs,Vars2,VarsF).

extr_list([],_,[]).
extr_list([H|T],S,SF):-
	member(H,S),
	! , extr_list(T,S,SF).
extr_list([H|T],S,[H|SF]):-
	extr_list(T,S,SF).

stop_comp(F,N,NP):-
	retract(selected_program(_)),
	retract(dt(_,F,N,DT)),
	rm_rl(DT),
	last_program(LP),
	retract(last_program(LP)),
	assertz(last_program(NP)).

rm_rl(rule(NR,_,_)):-
	retract(pair(_,rule(NR,_,_))),
	last_rule(LR),
	retract(last_rule(LR)),
	RR is LR-1,
	assert(last_rule(RR)).
rm_rl(branch(_,_,[DT])):-
	rm_rl(DT).

merge_er([],_,_,[]):-!.
merge_er([er(Subst,RhsS)|States],Lhs_ing,N,[rule(unfold(N),LhsS,RhsS)|Rules]):-
	substitute(Subst,Lhs_ing,LhsS),
	merge_er(States,Lhs_ing,N,Rules).


change_type_rule([],_,_).
change_type_rule([er(_,Rhs_ing_folded)|Sts_folded],NP3,NR_ing):-
	unf(NP3,F,rule(NF,Lhs_ing_folded,Rhs_ing_folded)),
	retract(unf(NP3,F,rule(NF,Lhs_ing_folded,Rhs_ing_folded))),
	assertz(fold(NP3,NR_ing,rule(NF,Lhs_ing_folded,Rhs_ing_folded))),
	change_type_rule(Sts_folded,NP3,NR_ing).

parser_ok(NewRule):- 
	( parser(NewRule,1),!
	; ( auto(no), write('    ERROR: Parser error with the rule '),nl
	  ; auto(yes)
	  ),
	  fail
	).

ctrl_folded(F_ing,N_ing,NP1,NR,Lhs_ed,Rhs_ed):-
	( NR<0,
	  ! , 
	  ( auto(no), nl,nl,write('    ERROR: Rule does not exist'),nl
	  ; auto(yes)
	  ),
	  stop_comp(F_ing,N_ing,NP1),fail
	; last_rule(LR),
	  ( NR = LR,
	    ! , 
	    ( auto(no),
	      ! ,nl,nl,write('    ERROR: Folded Rule '), write(NR),
	      write(' is the same as the Folding Rule'),nl
	    ; auto(yes)
	    ),
	    stop_comp(F_ing,N_ing,NP1),fail
	  ; NR > LR,
	    ! , 
	    ( auto(no),
	      ! , nl,nl,write('    ERROR: Rule not found'),nl
	    ; auto(yes)
	    ),
	    stop_comp(F_ing,N_ing,NP1),fail
	  )
	; % NR e' ok 
	  NR > 0
	),
	( unf(_,_,rule(NR,Lhs_ed,Rhs_ed)),
	  \+(elim(_,NR)), !
	; fold(_,Father,rule(NR,Lhs_ed,Rhs_ed)),
	  at_least_one(Father),
	  \+(elim(_,NR)), !
	; pair(_,rule(NR,Lhs_ed,Rhs_ed)),
	  is_in_R0(Lhs_ed), !
	; ( auto(no),
	    ! , nl,nl,write('    ERROR: Rule number '), write(NR),
	    write(' is in danger of sef-folding'),
	    stop_comp(F_ing,N_ing,NP1),fail
	  ; auto(yes)
	  )
	).

try_folding(Rhs_ed,Lhs_ing,Rhs_ing,NR,Rhs_ed_folded):-
	% Estrae le posizioni di Rhs_ed
	extract_positions(Rhs_ed,Positions),

	cont_try_f(Positions,1,Rhs_ed,Lhs_ing,Rhs_ing,_,Rhs_ed_folded),
	( \+(Rhs_ed_folded=Rhs_ed),\+(Rhs_ed_folded=inf),!
	; ( Rhs_ed_folded=Rhs_ed
	  ; Rhs_ed_folded=inf
	  ),
	  ( auto(no), nl,nl,write('    ERROR: in Folding Rule number '),
	    write(NR),nl
	  ; auto(yes)
	  ),
	  last_program(NP),NP1 is NP - 1, 
	  apply(f(F_ing),A)=Lhs_ing, num_args(A,N_ing),
	  stop_comp(F_ing,N_ing,NP1),
	  ! , fail
	).


elim_pos([],_,[]).
elim_pos([H|T],P,R):-
	append(P,_,H),
	!, elim_pos(T,P,R).
elim_pos([H|T],P,[H|R]):-
	elim_pos(T,P,R).

append_pos([], _, []).
append_pos([Pos|LPos], Pref, [NewPos|LPosOut]):-
	append(Pref,Pos,NewPos),
	append_pos(LPos,Pref,LPosOut).

show_c :- 
  nl, nl,
  nl,nl,write('************************************************************'),
  nl,   write('***                 RESULTS OF COMPOSITION               ***'),
  nl,   write('************************************************************'),
  nl,   write('***                                                      ***'),
  nl,   write('*** all   --> Show the resulted program                  ***'),
  nl,   write('***                                                      ***'),
  nl,   write('*** svall --> Save the resulted program                  ***'),
  nl,   write('***                                                      ***'),
  nl,   write('*** ret   --> Return to the main menu                    ***'),
  nl,   write('***                                                      ***'),
  nl,   write('************************************************************'),
  nl,   write('************************************************************'),
  nl,nl,write('    OPTION: '),read(Option),opt_c(Option).

opt_c(all) :- !,
	last_program(LP),nl,nl,write('NUMBER     RULE'),nl, 
	show_rules(LP),
	press,
	show_c.
opt_c(svall) :- !,
	nl,write('    Filename: '), read(File),
	tell(File),
	save_all,
	told,
	nl,write('    MESSAGE: File saved'),nl,
	press,
	show_c.
opt_c(ret) :- !.
opt_c(X) :- 
	nl,nl,write('    ERROR: Option '),write(X),write(' does not exist'),
	press,
	show_c.

substitute(S,v(N),Expr) :-
	memberchk((v(N),Expr),S), ! .

substitute(_,v(N),v(N)).

substitute(_,apply(F,[]),apply(F,[])) :- ! .

substitute(S,apply(F,Args),apply(F,ArgsS)) :-
	substitute_argsargs(S,Args,ArgsS).

substitute(S,[Cond,LocDefs],[CondS,LocDefsS]) :-
	substitute(S,Cond,CondS),
	substitute_LDs(LocDefs,S,LocDefsS).

substitute_argsargs(_,[],[]).

substitute_argsargs(S,[A|R],[AS|RS]) :-
	substitute(S,A,AS),
	substitute_argsargs(S,R,RS).


substitute_LDs([],_,[]).
substitute_LDs([apply(c("free"),X)|LDs],S,[apply(c("free"),X)|LDsS]):-
	! , substitute_LDs(LDs,S,LDsS).
substitute_LDs([LocDef|LDs],S,[LocDefS|LDsS]):-
	substitute_LocDef(S,LocDef,LocDefS),
	substitute_LDs(LDs,S,LDsS).

substitute_LocDef([],Rhs,Rhs).
substitute_LocDef([(v(X),_)|S],(apply(f(F),Args),Rhs),LocDefS):-
	is_present(apply(f(F),Args),v(X)),
	! , substitute_LocDef(S,(apply(f(F),Args),Rhs),LocDefS).

substitute_LocDef([(v(X),Exp)|S],(Lhs,Rhs),LocDefS):-
	! , substitute([(v(X),Exp)],Lhs,NewLhs),
	substitute([(v(X),Exp)],Rhs,NewRhs),
	substitute_LocDef(S,(NewLhs,NewRhs),LocDefS).


substitute_subs( [], Subst, Subst ).
substitute_subs( [Subst|LSubsts], (v(M),Term), SubRes ) :-
        substitute( Subst, Term, Term2 ),
        substitute_subs( LSubsts, (v(M),Term2), SubRes ).

substitute_SUBS( _, [], [] ).
substitute_SUBS( Subst, [(v(M),Term)|LSubsts], [NElSubst|NLSubsts] ) :-
        substitute_subs( Subst, (v(M),Term), NElSubst ),
        substitute_SUBS( Subst, LSubsts, NLSubsts ).

uniun_SUBS_st(er(NSubst,Rhs),Subst,er(Subst3,Rhs)):-
	substitute_SUBS(NSubst,Subst,Subst2),
	append(Subst2,NSubst,Subst3).

uniun_SUBS_stL([],_,[]):-!.
uniun_SUBS_stL([State|States],Subst,[ResState|Result]):-
	uniun_SUBS_st(State,Subst,ResState),
	uniun_SUBS_stL(States,Subst,Result).

is_present(X,X):- !.
is_present(apply(_,Args),X):- is_present_arg(Args,X).

is_present_arg([],_):- fail.
is_present_arg([Y|_],X):- is_present(Y,X).
is_present_arg([_|T],X):- is_present_arg(T,X).
	
pattern_match_rhs([Exp,_],Fix,Match) :-
	match_aux([(Exp,Fix)],[],Subs),
	( Subs == fallo, !, fail 
	; Match = Subs
	).

pattern_match(Exp,Fix,Match) :-
	match_aux([(Exp,Fix)],[],Subs),
	( Subs == fallo, !, fail
	; Match = Subs
	).
	
match_aux([],S,S).

match_aux([(v(N),Expr)|_],Subs,fallo) :-
	member((v(N),Term),Subs), Expr \== Term,!.
	
match_aux([(v(N),Expr)|Rest],Subs,SubsT) :-
	match_aux(Rest,[(v(N),Expr)|Subs],SubsT).

match_aux([((Exp1,[Cond,LocDefs]),(Exp2,[Cond2,LocDefs2]))|Rest],Subs,SubsT) :- 
	L1=(Exp1,Exp2),
	L2=(Cond,Cond2),
	zip_occ(LocDefs,LocDefs2,L3),
	! ,
	append(L3,Rest,Temp),
	Temp2 = [L1,L2|Temp],
	match_aux(Temp2,Subs,SubsT).

match_aux([((_,[_,_]),(_,[_,_]))|_],_,fallo).

match_aux([(apply(F,A),apply(F,B))|Rest],Subs,SubsT) :- 
	zip_occ(A,B,L1), !,
	append(L1,Rest,L2),
	match_aux(L2,Subs,SubsT).
	
match_aux([(apply(_,_),apply(_,_))|_],_,fallo).

substitute_expr(_,[],Expr1,Expr1).

substitute_expr(apply(F,Args),[M|L],Expr1,apply(F,Expr2)) :-	% in apply qlsi
	substitute_at(M,Args,L,Expr1,Expr2).

substitute_expr([_,LocDs],[1], SubExpr,[SubExpr,LocDs]):-!. 	% in rhs
substitute_expr([Cond,LocDs],[1|ROcc], SubExpr,[NewCond,LocDs]):-
	substitute_expr(Cond,ROcc,SubExpr,NewCond).
substitute_expr([Cond,_],[2], SubExpr,[Cond,SubExpr]):-!.
substitute_expr([Cond,LocDefs],[2,N|ROcc], SubExpr,[Cond,LocDefsOut]):-
	substitute_at(N,LocDefs,ROcc,SubExpr,LocDefsOut).

substitute_at(1,[Arg1|T],L,Exp,[NArg1|T]) :-
	! , substitute_expr(Arg1,L,Exp,NArg1).
	
substitute_at(N,[H|T],L,Exp,[H|NT]) :-
	N>1, M is N-1,
	substitute_at(M,T,L,Exp,NT).


subexpression(E,[],E):-!.

subexpression(apply(_,Args),[N|L],E) :-
	!, select_arg(Args,1,N,ArgN),
	subexpression(ArgN,L,E).

subexpression([Cond,_],[1],Cond):-!.
subexpression([Cond,_],[1|ROcc],NewCond):-
	subexpression(Cond,ROcc,NewCond).
subexpression([_,LocDs],[2],LocDs):-!.
subexpression([_,LocDefs],[2,N|ROcc],LocDefOut):-
	select_arg(LocDefs,1,N,LocDefN),
	subexpression(LocDefN,ROcc,LocDefOut).

select_arg([E|_],Max,Max,E):-!.

select_arg([_|R],M,Max,A) :-
	M<Max,
	! , NM is M+1,
	select_arg(R,NM,Max,A).

last_prg_vble("A"). % this is to produce fresh vbles for programs

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%            GENERATION OF DEFINITIONAL TREES                     %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


strategy(optmatch).

basicDefTrees :-
	assert( dt(0,"ifelse", 3,
	  branch( apply(f("ifelse"),[ v("X1"), v("X2"), v("X3") ]),[1],[
	    rule(0,apply(f("ifelse"),[apply(c("true"),[]),v("X2"),v("X3")]),
								[v("X2"),[]] ),
	    rule(0,apply(f("ifelse"),[apply(c("false"),[]),v("X2"),v("X3")]),
							[v("X3"),[]]) ] ) ) ),
	assert( dt(0,"if", 3,
	  branch( apply(f("if"),[ v("X1"), v("X2") ]),[1],[
	    rule(0,apply(f("if"),[apply(c("true"),[]),v("X2")]),
							[v("X2"),[]] ) ] ) ) ).

deftrees :-
        retractall( dt(_,_,_,_) ), basicDefTrees,
	last_rule(Max), M is Max +1,
        partition( 1, M, [], OrderedRules ),
        ( auto(no), ! , nl, write('Generating definitional trees... ')
	; auto(yes)
	),
        gt_process( OrderedRules ),
        list_deftrees(0).

% -- Partition rules by function symbol
partition( Max, Max, _,[] ):- ! .
partition( NR, Max, AccOper, [SameRules|Rules] ) :-
	look_for_rule(0,NR,Lhs,Rhs),
	apply(f(F),Args) = Lhs,
	\+(member(F,AccOper)),
	! , AccOper2 = [F|AccOper],
	num_args(Args, NArgs),
	M is NR+1, 
	( (M < Max, 
           ! , sameOper( F, NArgs, M, Max, SameRules2 ),
           SameRules = [ rule(NR,Lhs,Rhs) | SameRules2 ],
           partition( M, Max, AccOper2, Rules ) ) 
	 ;(SameRules = [ rule(NR,Lhs,Rhs) ],
	   Rules=[])
	) .

partition( N, Max, AccOper, Rules ) :- % regola N gia' considerata
	look_for_rule(0,N,Lhs,_),
	apply(f(F),_) = Lhs,
	member(F,AccOper) ,
	K is N+1 ,
        partition( K, Max, AccOper, Rules ).

sameOper( _, _, Max, Max, [] ):- !.
sameOper( F, NA, R, Max,[rule(R,Lhs,Rhs) | SameRules ] ) :- 
	look_for_rule(0,R,Lhs,Rhs),
	apply(f(F),Args) = Lhs,
	num_args(Args, NA),
	!,
	M is R+1, 
        sameOper( F, NA, M, Max, SameRules).
sameOper( F, NA, R, Max, SameRules) :- 
	M is R+1,
        sameOper( F, NA, M, Max, SameRules).

% -- Auxiliar for gt
gt_process( [] ).
gt_process( [SameRules|Rules] ) :-
        [ rule(_, apply(f(F),Args), _ ) |_] = SameRules,
	num_args(Args,N),
	( N>0 ,
	  last_prg_vble(W),
	  list_vars(1, N, W, Vars),
	  update_prg_vbles
	; Vars = []
	),
	( dt(_,F,N,_), 
	  ( auto(yes),!,fail
	  ; auto(no),
	    format('ERROR: Previous definitional tree for ~s\n',[F]),
	    ! , fail 
	  )
	; strategy( Strategy )
	),
        ( gt( apply(f(F),Vars), Strategy, SameRules, DefTree ), 
	  ! 
        ; ( auto(yes),!,fail
	  ; auto(no),
	    format('ERROR: No definitional tree cannot function ~s\n',[F]),
	    !, fail 
	  )
	),
	last_program(LP),
        assert( dt(LP, F, N, DefTree ) ), 
        ! , gt_process( Rules ).

list_vars(Max, Max, W,[v([88,C|W])] ):- 
	! , C is Max+"0".
list_vars(N, Max, W, [v([88,C|W])|Vars] ):-
	N1 is N+1, C is N+"0",
	list_vars(N1, Max, W, Vars).

list_deftrees(NP) :- dt(NP,F,_,_), 
        F \== "if",F \== "ifelse",
        format("~s ",F), fail.
list_deftrees(_) :- nl.

% -- Get demanded occurrencies --------------------------------------------
dp( apply(f(_),Terms), ListRules, Positions ) :- 
        var_pos( Terms, [], 1, [], VarPos ),
        dp( ListRules, VarPos, ListRules, Positions ).

var_pos( [], _, _, Vars, Vars ).
var_pos( [v(_)|Terms], Ox, N, Vars1, Vars3 ) :- 
        N1 is N + 1,
        append( Ox, [N], Ox1 ),
        append( Vars1, [Ox1], Vars2 ),
        var_pos( Terms, Ox, N1, Vars2, Vars3 ).
var_pos( [apply(c(_),Args)|Terms], Ox, N, Vars1, Vars3 ) :-
        N1 is N + 1,
        append( Ox, [N], Ox1 ),
        var_pos( Args, Ox1, 1, Vars1, Vars2 ),
        var_pos( Terms, Ox, N1, Vars2, Vars3 ).

dp( _, [], _, [] ). % -- no more 
dp( [], [_|Ox], ListRules, Positions ) :- % -- no more rules
        dp( ListRules, Ox, ListRules, Positions ).
dp( [rule(_,L,_)|_], [O|Ox], ListRules, [O|Positions] ) :- % -- found constructor
        subexpression( L, O, apply(c(_),_) ),
        dp( ListRules, Ox, ListRules, Positions ).
dp( [rule(_,L,_)|Rules], [O|Ox], ListRules, Positions ) :- % -- found variable
        subexpression( L, O, v(_) ),
        dp( Rules, [O|Ox], ListRules, Positions ).

% -- Main predicate for Definitional Tree ------------------------------

% ---- branch node ( all constructors )

gt( apply(f(F),Terms), optmatch, Rules, Tree ) :- 
        dp( apply(f(F),Terms), Rules, [P|Positions] ),
        ( find_opt( [P|Positions], Rules, O ),!
         ;O=P ),
        groups( Rules, O, Grouped ),
        Tree = branch(  apply(f(F),Terms), O,  SubTrees ),
        subtrees( Grouped, apply(f(F),Terms), O, optmatch, SubTrees ).

find_opt( [V|_], Rules, V ) :-
        split( Rules, V, Rules, [] ).
find_opt( [V|Vs], Rules, O ) :-
        \+(split( Rules, V, Rules, [] )),
        find_opt( Vs, Rules, O ).        

% ----- branch node ( all constructors )
gt( apply(f(F),Terms), leftright, Rules, Tree ) :- 
        dp( apply(f(F),Terms), Rules, [O|_] ),
        split( Rules, O, Rules, [] ),
        groups( Rules, O, Grouped ),
        Tree = branch( apply(f(F),Terms), O, SubTrees ),
        subtrees( Grouped, apply(f(F),Terms), O, leftright, SubTrees ).

subtrees( [], _, _, _, [] ).
subtrees( [R|Rs], Term, O, Strategy, [SubTree|SubTrees] ) :-
        [rule(_, L, _ )|_] = R, 
        subexpression( L, O, apply(c(F),Args) ),
	num_args(Args,N),
	( ( N>0 ,
	    last_prg_vble(W),
	    list_vars(1, N, W, Vars),
	    update_prg_vbles )
	 ;Vars = []
	),
        substitute_expr( Term, O, apply(c(F),Vars), TermS ),
        gt( TermS,Strategy, R, SubTree ),
        subtrees( Rs, Term, O, Strategy, SubTrees ).

% -------- ( constructors and variables )
gt( apply(f(F),Terms), _, Rules, _ ) :- 
        dp( apply(f(F),Terms), Rules, [O|_] ),
        split( Rules, O, CRules, VRules ), CRules \== [], VRules \== [],
 	( auto(yes), ! , fail
	; auto(no),
	  format('ERROR: No definitional tree cannot function ~s\n',[F]),
	  !, fail
	).

% -------- rule node
gt( apply(f(F),Terms), _, Rules, Tree ) :- 
	! , dp( apply(f(F),Terms), Rules, [] ),
	( Rules=[rule(NR, Lhs,[R,LD])],
	  ! , Tree=rule(NR, Lhs,[R,LD])
 	; ( auto(yes), ! , fail
	  ; auto(no), nl, 
pippo,
	    !,nl,
	    format('ERROR: "~s" not inductively sequential operation\n',[F]),
	    fail
	  )
	).
pippo.

% -- Split rules between constructors and variables -----------------
split( Rules, Ox, RCons, RVars ) :- split( Rules, Ox, [], [], RCons, RVars ).

split( [], _, RCons, RVars, RCons, RVars ).
split( [rule(NR,L,R)|Rules], Ox, RCons, RVars, RCons3, RVars2 ) :-
        subexpression( L, Ox, apply(c(_),_) ),
        append( RCons, [rule(NR,L,R)], RCons2 ),
        split( Rules, Ox, RCons2, RVars, RCons3, RVars2 ).
split( [rule(NR,L,R)|Rules], Ox, RCons, RVars, RCons2, RVars3 ) :-
        subexpression( L, Ox, v(_) ),
        append( RVars, [rule(NR,L,R)], RVars2 ),
        split( Rules, Ox, RCons, RVars2, RCons2, RVars3 ).

% -- Group rules by same constructor -------------------------------
groups( [], _, [] ).
groups( [rule(NR,L,R)|Rules], Ox, [SameRules|Rules3] ) :-
        subexpression( L, Ox, apply(c(F),Args) ),
	num_args(Args, NArgs),
        extract( Rules, Ox, F, NArgs, [], SameRules1, Rules2 ), 
        SameRules = [rule(NR,L,R)|SameRules1],
        groups( Rules2, Ox, Rules3 ).

extract( [], _, _, _, SameRules, SameRules, [] ).
extract( [rule(NR,L,R)|Rules], Ox, F, N, AccSameRules, SameRules, Rules2 ) :-
        subexpression( L, Ox, apply(c(F),Args) ),	% -- included
	num_args(Args,N),
        append( AccSameRules, [rule(NR,L,R)], AccSameRules2 ),
        extract( Rules, Ox, F, N, AccSameRules2, SameRules, Rules2 ).
extract( [rule(NR,L,R)|Rules], Ox, F, N, AccSameRules, SameRules, 
         [rule(NR,L,R)|Rules2] ) :-
        subexpression( L, Ox, apply(c(F1),Args) ),
	num_args(Args,N1),
	(F1 \== F ; N1 =\= N ),  % -- non included
        extract( Rules, Ox, F, N, AccSameRules, SameRules, Rules2 ).


build_unf([]).
build_unf([rule(unfold(NR),apply(f(F),Args),Rhs)|Rules]):-
	num_args(Args,N),
	selected_program(Prog),
	find_dt(Prog,F,N,DT),
	search_rl_dt(DT,NR,[rule(unfold(NR),apply(f(F),Args),Rhs)|Rules],NewDT),
	retract(last_program(LP)),
	NP is LP+1,
	assert(last_program(NP)),
	assert(dt(NP,F,N,NewDT)).

search_rl_dt(rule(NRule,Lhs,_),NRule,Rules,NDT):-
	! ,
	strategy( Strategy ),
	gt( Lhs, Strategy, Rules, DT),	
	ins_rl_from_dt(DT,NDT).
search_rl_dt(rule(NR,_,_),NRule,_,_):-
	! , NRule \== NR, fail.
search_rl_dt(branch(P,O,Dts),NRule,Rules,branch(P,O,NDts)):-
	! , search_rl_dtL(Dts,NRule,Rules,NDts).
search_rl_dtL([],_,_,_) :- fail.
search_rl_dtL([DT|DTs],NRule,Rules,NDTs):-
	( search_rl_dt(DT,NRule,Rules,NDT), 
	  ! , NDTs= [NDT|DTs]
	; search_rl_dtL(DTs,NRule,Rules,DTs2),
	  ! , NDTs= [DT|DTs2]
	).

ins_rl_from_dt(rule(unfold(Father),Lhs,Rhs),rule(NR,Lhs,Rhs)):-
	! , insert_new_rule(unf,Father,Lhs,Rhs,NR).
ins_rl_from_dt(rule(fold(Father),Lhs,Rhs),rule(NR,Lhs,Rhs)):-
	! , insert_new_rule(fold,Father,Lhs,Rhs,NR).
ins_rl_from_dt(rule(intr(N),Lhs,Rhs),rule(NR,Lhs,Rhs)):-
	! , insert_new_rule(pair, N, Lhs, Rhs, NR).
ins_rl_from_dt(branch(P,O,Dts),branch(P,O,NDts)):-
	! , ins_rl_from_dtL(Dts,NDts).
ins_rl_from_dtL([],[]).
ins_rl_from_dtL([DT|DTs],NDTs):-
	ins_rl_from_dt(DT,NDT), 
	ins_rl_from_dtL(DTs,DTs2),
	NDTs= [NDT|DTs2].

build_intr([]).
build_intr(Rules):-
	analyze(Rules,NewRules),
	partition_rl(NewRules,OrderedRules),
        ( auto(no), ! , nl, write('Generating definitional trees... ')
	; auto(yes)
	),
	last_program(NP),
	retract(last_program(NP)),
	LP is NP + 1 ,
	assert(last_program(LP)),
        gt_process( OrderedRules ),
	( dt(LP,_,_,_), 
	  ! , 
	  retract(last_program(LP)),
	  assert(last_program(NP)),
	  update_dt(LP),
	  retract(last_program(NP)),
	  assert(last_program(LP)),
	  ( auto(no), !, list_deftrees(LP)
	  ; auto(yes)
	  )
	; retract(last_program(LP)),
	  assert(last_program(NP))
	).

update_dt(LP):-
	dt(LP,F,N,DT),
	ins_rl_from_dt(DT,NDT),
	retract(dt(LP,F,N,DT)),
	assert(dt(LP,F,N,NDT)), 
	fail.
update_dt(_).

partition_rl([], [] ).
partition_rl([rule(NR,apply(f(F),Terms),Rhs)|Rules1],[SameRules|Rules3]) :-
	num_args(Terms,N),
        sameOper_rl( F, N, Rules1, SameRules1, Rules2 ),
        SameRules = [rule(NR,apply(f(F),Terms),Rhs) | SameRules1 ],
        partition_rl( Rules2, Rules3 ).

sameOper_rl( _, _, [], [], [] ).
sameOper_rl( F, N,
		[rule(NR,apply(f(F),Terms),Rhs)|Rules1], 
		[rule(NR,apply(f(F),Terms),Rhs)|MRules2], Rules2 ) :- 
	num_args(Terms,N),
	! , sameOper_rl( F, N, Rules1, MRules2, Rules2 ).
sameOper_rl( F, N, [rule(NR,apply(f(F1),Terms),Rhs)|Rules1], 
		MRules2, [rule(NR,apply(f(F1),Terms),Rhs)|Rules2]) :-
	sameOper_rl( F, N, Rules1, MRules2, Rules2 ).

del_rl_dt(rule(NR,_,_)):-
	last_program(LP), NP is LP+1,
	assertz(elim(NP,NR)).
del_rl_dt(branch(_,_,Rules)):-
	del_rl_dtL(Rules).
del_rl_dtL([]).
del_rl_dtL([Rule|Rules]):-
	del_rl_dt(Rule),
	del_rl_dtL(Rules).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%                          NARROWING                              %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


narrow_step_Rhs([Exp,LD],Result):-
	eval(Exp,Res),
	! , control_ld(Res,LD,Result).

control_ld([],_,[]).
control_ld([er(Sust,[Exp,NewLD])|Res],LD,[er(Sust,[Exp,LD3])|NRes]):-
	substitute_LDs(LD,Sust,LD2),
	append(NewLD,LD2,LD3),
	control_ld(Res,LD,NRes).

% eval -------------------------------------------------------------------
% -- Special: constraint conjuction
eval( apply(f("&"),[apply(c("success"),[]),Exp2]), States ) :-
	States = [ er( [], [Exp2,[]] ) ].
eval( apply(f("&"),[Exp1,apply(c("success"),[])]), States ) :-
	States = [ er( [], [Exp1,[]] ) ].
eval( apply(f("&"),[Exp1,Exp2]), States ) :-
	eval( Exp1, StatesC ),
	replace( apply(f("&"),[Exp1,Exp2]),[1], StatesC, States ).
eval( apply(f("&"),[Exp1,Exp2]), States ) :-
	eval( Exp2, StatesC ),
	replace( apply(f("&"),[Exp1,Exp2]),[2], StatesC, States ).

% -- Special: Strict Equality
eval( apply(f("=="),[ Term1, Term2 ]), States ) :- !,
	( % Case Constructors
	  ( Term1 = apply(c(F),[]), Term2 = apply(c(F),[]),
	    States = [ er( [], [ apply(c("true"),[]) ,[]] ) ]
	  ; Term1 = apply(c(F),Args1), Term2 = apply(c(F),Args2),
	    num_args(Args1,N),num_args(Args2,N),
	    eval_equality( Args1, Args2, Term3 ), 
	    States = [ er( [], [Term3,[]] ) ]
	   ;Term1 = apply(c(F1),Args1), Term2 = apply(c(F2),Args2), 
	    num_args(Args1,N1),num_args(Args2,N2),
	    (F1\==F2;N1=\=N2),
	    States = [ er( [], [apply(c("false"),[]), []] ) ] 
	  )
	 ;% Case Function
	  ( Term1 = apply(f(F),Terms), Ox = [1]
	   ;Term2 = apply(f(F),Terms), Ox = [2] 
	  ), 
	  ! , eval( apply(f(F),Terms), StatesF ),
	  replace( apply(f("=="),[Term1,Term2]), Ox, StatesF, States )
	).

eval_equality( [X], [Y], apply(f("=="),[ X, Y ]) ).
eval_equality( [X1,X2|Xs], [Y1,Y2|Ys], 
               apply(f("&&"),[ apply(f("=="),[X1,Y1]),Rest]) ) :-
        eval_equality([X2|Xs],[Y2|Ys],Rest).

% -- Special: Arithmetic functions
eval( apply(f("+"),[Term1,Term2]), States ) :- !, 
        eval_Op( "+", Term1, Term2, States ).
eval( apply(f("-"),[ Term1, Term2 ]), States ) :- !, 
        eval_Op( "-", Term1, Term2, States ).
eval( apply(f("*"),[ Term1, Term2 ]), States ) :- !, 
        eval_Op( "*", Term1, Term2, States ).
eval( apply(f("'div'"),[ Term1, Term2 ]), States ) :- !, 
        eval_Op( "'div'", Term1, Term2, States ).
eval( apply(f("'mod'"),[ Term1, Term2 ]), States ) :- !, 
        eval_Op( "'mod'", Term1, Term2, States ).
eval( apply(f("%"),[ Term1, Term2 ]), States ) :- !, 
        eval_Op( "%", Term1, Term2, States ).
eval( apply(f(">"),[ Term1, Term2 ]), States ) :- !, 
        eval_Op( ">", Term1, Term2, States ).
eval( apply(f("<"),[ Term1, Term2 ]), States ) :- !, 
        eval_Op( "<", Term1, Term2, States ).
eval( apply(f(">="),[ Term1, Term2 ]), States ) :- !, 
        eval_Op( ">=", Term1, Term2, States ).
eval( apply(f("=<"),[ Term1, Term2 ]), States ) :- !, 
        eval_Op( "=<", Term1, Term2, States ).

% common behaviour
eval_Op( Op, Term1, Term2, States ) :-
        (  Term1 = apply(c(ANum1),[]), Term2 = apply(c(ANum2),[]), 
	   ! , number_chars(Num1,ANum1), number_chars(Num2,ANum2),
	   eval_OpValue( Op, Num1, Num2, Num3 ),
	   ( ( Num3=="true" 
	     ; Num3=="false"
	     ), Num3=ANum3 
	   ; number_chars(Num3,ANum3)
	   ),
           States = [ er( [], [apply(c(ANum3),[]), []] ) ]

         ; (  Term1 = apply(f(F),Terms), Ox = [1]
            ; Term2 = apply(f(F),Terms), Ox = [2] 
	   ), 
	   ! , eval( apply(f(F),Terms), StatesF ),
           replace( apply(f(Op),[Term1,Term2]), Ox, StatesF, States )  
        ).

eval_OpValue( "+", Num1, Num2, Num3 ) :- 
        Num3 is Num1 + Num2.
eval_OpValue( "-", Num1, Num2, Num3 ) :- 
        Num3 is Num1 - Num2.
eval_OpValue( "*", Num1, Num2, Num3 ) :- 
        Num3 is Num1 * Num2.
eval_OpValue( "'div'", Num1, Num2, Num3 ) :- 
        Num3 is Num1 // Num2.
eval_OpValue( "'mod'", Num1, Num2, Num3 ) :- 
        Num3 is Num1 mod Num2.
eval_OpValue( ">", Num1, Num2, Num3 ) :- 
        Num1 > Num2, !, Num3 = "true" ; Num3 = "false".
eval_OpValue( "<", Num1, Num2, Num3 ) :- 
        Num1 < Num2, !, Num3 = "true" ; Num3 = "false".
eval_OpValue( ">=", Num1, Num2, Num3 ) :- 
        Num1 >= Num2, !, Num3 = "true" ; Num3 = "false".
eval_OpValue( "=<", Num1, Num2, Num3 ) :- 
        Num1 =< Num2, !, Num3 = "true" ; Num3 = "false".

% ----------- Comparisson X=Y 
% --- functions ------
eval( apply(f("=:="),[apply(f(F),Terms), Term2 ]), States ) :- 
        eval( apply(f(F),Terms), StatesT ),
        replace( apply(f("=:="),[ apply(f(F),Terms), Term2 ]), [1], StatesT, States ).

% --- functions ------
eval( apply(f("=:="),[ Term1, apply(f(F),Terms) ]), States ) :- 
        eval( apply(f(F),Terms), StatesT ),
        replace( apply(f("=:="),[ Term1, apply(f(F),Terms) ]), [2], StatesT, States ).

% --- same elementary constructor -
eval( apply(f("=:="),[ apply(c(F),[]), apply(c(F),[]) ]), States ) :- 
        States = [er( [], [apply(c("success"),[]), []] )].

% --- same constructor -
eval( apply(f("=:="),[ apply(c(F),Terms1), apply(c(F),Terms2) ]), States ) :- 
	num_args(Terms1,N1),num_args(Terms2,N2),
        ( N1=N2,
	  States = [er( [], [Constraints, []] )],
          eval_equations( Terms1, Terms2, Constraints )
	; States=[]
	).

% --- different constructor ---
eval( apply(f("=:="),[ apply(c(F1),_), apply(c(F2),_) ]), [] ) :- 
	        F1 \== F2 .

% --- variable --------------
eval( apply(f("=:="),[ v(N), v(M) ]), States ) :- 
        States = [er( [( v(N), v(M) )], [apply(c("success"),[]), []] )].

% --- variable and elementary contructor ------
eval( apply(f("=:="),[ v(M), apply(c(F),[]) ]), States ) :- 
        !,States = [er([(v(M),apply(c(F),[]) )], [apply(c("success"),[]),[]] )].

% --- variable and elementary contructor ------
eval( apply(f("=:="),[ apply(c(F),[]), v(M) ]), States ) :- 
        !,States = [er([(v(M),apply(c(F),[]) )], [apply(c("success"),[]),[]] )].

% --- variable and contructor ------
eval( apply(f("=:="),[ v(M), apply(c(F),Terms) ]), States ) :- 
        \+( occur_check( M, apply(c(F),Terms) ) ),
	num_args(Terms,N),
	last_prg_vble(W),
	list_vars(1, N, W, Vars),
	update_prg_vbles,
        eval_equations( Vars, Terms, Constraints ),
        States = [er( [( v(M), apply(c(F),Vars) )], [Constraints, []] )].

% --- variable and contructor ------
eval( apply(f("=:="),[ apply(c(F),Terms), v(M) ]), States ) :- 
        \+( occur_check( M, apply(c(F),Terms) ) ),
	num_args(Terms,N),
	last_prg_vble(W),
	list_vars(1, N, W, Vars),
	update_prg_vbles,
        eval_equations( Vars, Terms, Constraints ),
        States = [er( [( v(M), apply(c(F),Vars) )], [Constraints, []] )].

eval( apply(f("=:="),[ v(M), apply(c(F),Terms) ]), [] ) :-
	occur_check( M, apply(c(F),Terms) ).	% --- occur_check
eval( apply(f("=:="),[ apply(c(F),Terms), v(M) ]), [] ) :-
	occur_check( M, apply(c(F),Terms) ). 	% --- occur_check

eval_equations([Term1|Tx1],[Term2|Tx2],Ct3 ) :-
        eval_eqs( Tx1, Tx2, apply(f("=:="),[Term1,Term2]), Ct3 ).

eval_eqs( [], [], Constraints, Constraints ).
eval_eqs([Term1|Tx1],[Term2|Tx2],Ct2,apply(f("&"),[Ct2,Ct3]) ) :-
        eval_eqs( Tx1, Tx2, apply(f("=:="),[Term1,Term2]), Ct3 ).

occur_check( M, v(M) ).
occur_check( M, apply(c(_),Terms) ) :- occur_checkL( M, Terms ).

occur_checkL( M, [Term|_] ) :- occur_check( M, Term ),!.
occur_checkL( M, [_|Terms] ) :- occur_checkL( M, Terms ).

% -- Find some evaluable subexpression
eval_some( [Term|_], Exp, N, States ) :-
        eval( Term, StatesT ),!,
        replace( Exp, [N], StatesT, States ).
eval_some( [_|Terms], Exp, N, States ) :-
        N1 is N + 1,
        eval_some( Terms, Exp, N1, States ).

% -- Main ---------------------------------------
eval( apply(f(F),Terms), States ) :- 
        num_args(Terms,N),
	selected_program(Prog),
	( find_dt(Prog, F, N, DefTree ),
	  last_prg_vble(M),
	  stand_apart(M, DefTree, ADefFRESH ),
	  update_prg_vbles,
	  evalT( apply(f(F),Terms), ADefFRESH, States ),!
	; States=[]
	).
eval( apply(c(F),Terms), States ) :- 
	eval_some( Terms, apply(c(F),Terms), 1, States ).

% -- Main for Definitional Trees -------------------------------
evalT( Expression, rule(_,TermL,TermR), [er([],Term)] ) :-    % -- Rule
        pattern_match( TermL, Expression, Susts ),
        substitute( Susts, TermR, Term ).
evalT( Expression, branch( _, Ox, Trees ), States ) :-
        subexpression( Expression, Ox, apply(c(F),Terms) ),
        evalT_some( Trees, Expression, Ox, apply(c(F),Terms), States ).
evalT( Expression, branch( _, Ox, Trees ), States ) :-
        subexpression( Expression, Ox, v(N) ),
        evalT_instantiate( Trees, Expression, Ox, N, States ).
evalT( Expression, branch( _, Ox, _ ), States ) :-
        subexpression( Expression, Ox, apply(f(F),Terms) ),
        eval( apply(f(F),Terms), StatesF ),
        replace( Expression, Ox, StatesF, States ).

evalT_some( [], _, _, apply(c(_),_), [] ). % -- error no one evaluable
evalT_some( [Tree|_], Expression, Occurrence, apply(c(F),_), States ) :-
        pattern( Tree, Pattern ),
	subexpression( Pattern, Occurrence, apply(c(F),_) ),
        evalT( Expression, Tree, States ),!.
evalT_some( [_|Trees], Expression, Occurrence, apply(c(F),_), States ) :-
        evalT_some( Trees, Expression, Occurrence, apply(c(F),_), States ).

evalT_instantiate( [], _, _, _, [] ).
evalT_instantiate( [Tree|Trees], Expression, Occurrence, N, StatesF ) :-
        pattern( Tree, Pattern ), 
        subexpression( Pattern, Occurrence, TermS ),
        Sust = (v(N),TermS), 
	substitute( [Sust], Expression, ExpressionS ),
	evalT(ExpressionS,Tree,StateC),
	join_state(StateC,Sust,State),
	append(State,States,StatesF),
        evalT_instantiate( Trees, Expression, Occurrence, N, States ).

join_state( [], _, [] ).
join_state( [er( Susts1, Exp )|States1], Susts, 
                [er( Susts3, Exp )|States2] ) :-
        substitute_subs( Susts1, Susts, Susts2 ),
        append( [Susts2], Susts1, Susts3 ),
        join_state( States1, Susts, States2 ).


pattern( branch( Pattern, _, _ ), Pattern ).
pattern( rule(_, Pattern, _), Pattern ).

find_dt(Prog, F, N, DefTree ):-
	find_dt(Prog, Prog, F, N, DefTree ).

find_dt(Prog, MaxP, F, N, DefTree ):-
	dt(Prog, F, N, DefTree ),
	! , 
	not_eliminated(Prog,MaxP,F,N).
find_dt(0, _, _, _, _ ):- ! , fail.
find_dt(P, MaxP, F, N, DefTree ):-
	! , P1 is P -1,
	find_dt(P1, MaxP, F, N, DefTree ).

look_for_dt(Prog, F, N, DefTree ):-
	dt(Prog, F, N, DefTree ), ! .
look_for_dt(0, _, _, _ ):- ! , fail.
look_for_dt(Prog, F, N, DefTree ):-
	! , P1 is Prog -1,
	look_for_dt(P1, F, N, DefTree ).

not_eliminated(Prog,_,F,N):-
	elim_dt(Prog,F,N), ! , fail.	
not_eliminated(LP,LP,_,_):- ! .		
not_eliminated(Prog,LP,F,N):-
	P is Prog + 1,
	not_eliminated(P,LP,F,N).

% replace --------------------------------------------------------
replace( _, _, [], [] ).
replace( Expression, Occurrence, 
         [er(Susts,[Exp,LD])|States], [er(Susts,[ExpR,LD])|StatesR] ) :-
        substitute( Susts, Expression, ExpressionS ),
        substitute_expr( ExpressionS, Occurrence, Exp, ExpR ),
        replace( Expression, Occurrence, States, StatesR ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%                      STANDARDIZATION APART                      %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

stand_apart( N, rule(NR, TermL, TermR ), rule(NR, TermLx, TermRx ) ) :-
        stand_apart( N, TermL, TermLx),
        stand_apart( N, TermR, TermRx).
stand_apart( N, branch( Pattern1, Ox, Trees1 ), 
		branch( Pattern2, Ox, Trees2 ) ) :-
        stand_apart(N, Pattern1,Pattern2 ),
        stand_apartL(N, Trees1, Trees2 ).

stand_apart(N, apply(F,Terms1), apply(F,Terms2) ) :-
        stand_apartL(N, Terms1, Terms2 ).
stand_apart(N, v(N1), v(N2) ) :- append(N1,N,N2).
stand_apart(N,[Cond,LocDefs],[NewCond,NewLocDefs]) :-
	stand_apart(N,Cond,NewCond),
	stand_apartL(N,LocDefs,NewLocDefs).
stand_apart(N,(Lhs,Rhs),(NewLhs,NewRhs)) :-
	stand_apart(N,Lhs,NewLhs),
	stand_apart(N,Rhs,NewRhs).

stand_apartL(_, [], [] ).
stand_apartL(N, [Term1|Terms1], [Term2|Terms2] ) :-
        stand_apart(N, Term1, Term2 ),
        stand_apartL(N, Terms1, Terms2 ).
% --------
update_prg_vbles :-
	retract(last_prg_vble(S)),!,
	length(S,LS),
	nth(LS,S,Last,Rest),
	(Last < 90, !, New is Last+1, append(Rest,[New],L);
	 append(S,[65],L)),
	assert(last_prg_vble(L)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	
?- menu.
